from collections import OrderedDict
from datetime import timedelta
from typing import Optional, Dict, Any, List, Tuple, AsyncIterable

from dropland.blocks.redis import RedisStorageEngine
from dropland.data.cache import ModelCacheProtocol, ModelCacheData
from dropland.data.context import ContextData
from dropland.data.serializers import Serializer, Deserializer


class SimpleRedisModelCache(ModelCacheProtocol):
    def __init__(self, engine: RedisStorageEngine, class_key: str,
                 serializer: Serializer, deserializer: Deserializer, ttl_enabled: bool = True):
        self._redis_engine = engine
        self._class_key = class_key
        self._serializer = serializer
        self._deserializer = deserializer
        self._ttl_enabled = ttl_enabled
        self._null_data = self._serializer.serialize(None)

    def get_model_cache_key(self) -> str:
        return f'{self._redis_engine.name}.models.{self._class_key}'

    async def cache_one(
            self, ctx: ContextData, data: ModelCacheData, ttl: Optional[timedelta] = None, **kwargs) -> bool:
        cache_kwargs = dict()

        if self._ttl_enabled:
            expiration = ttl or self._redis_engine.default_ttl
            total_seconds = expiration.total_seconds()

            if total_seconds > 1:
                cache_kwargs['expire'] = int(total_seconds)
            elif 0 < total_seconds < 1:
                cache_kwargs['pexpire'] = int(total_seconds * 1000)
            else:
                total_seconds = self._redis_engine.default_ttl.total_seconds()
                cache_kwargs['expire'] = int(total_seconds) if total_seconds > 1 else 60

        serialized = self._serializer.serialize(data.data)
        return bool(await ctx.redis.connection.set(self.get_cache_key(data.cache_id), serialized, **cache_kwargs))

    async def cache_many(
            self, ctx: ContextData, objects: List[ModelCacheData],
            ttl: Optional[timedelta] = None, **kwargs) -> bool:
        if not objects:
            return False
        res = False

        tx = ctx.redis.connection.multi_exec()

        for instance in objects:
            cache_kwargs = dict()

            if self._ttl_enabled:
                expiration = ttl or self._redis_engine.default_ttl
                total_seconds = expiration.total_seconds()

                if total_seconds > 1:
                    cache_kwargs['expire'] = int(total_seconds)
                elif 0 < total_seconds < 1:
                    cache_kwargs['pexpire'] = int(total_seconds * 1000)
                else:
                    total_seconds = self._redis_engine.default_ttl.total_seconds()
                    cache_kwargs['expire'] = int(total_seconds) if total_seconds > 1 else 60

            serialized = self._serializer.serialize(instance.data)
            tx.set(self.get_cache_key(instance.cache_id), serialized, **cache_kwargs)

        for r in await tx.execute():
            res |= r

        return bool(res)

    async def load_one(
            self, ctx: ContextData, cache_key: str, **kwargs) -> Tuple[bool, Optional[Any]]:
        if res := await ctx.redis.connection.get(cache_key):
            return True, self._deserializer.deserialize(res)
        return False, None

    async def load_many(
            self, ctx: ContextData, indices: List[Any], **kwargs) -> List[Optional[Dict[str, Any]]]:
        if not indices:
            return []

        cache_keys = [self.get_cache_key(id_value) for id_value in indices]
        res = await ctx.redis.connection.mget(*cache_keys)
        objects: Dict[Any, Any] = OrderedDict()

        for id_value, data in zip(indices, res):
            objects[id_value] = self._deserializer.deserialize(data) if data is not None else None

        return list(objects.values())

    async def drop_one(self, ctx: ContextData, cache_key: str) -> bool:
        return bool(await ctx.redis.connection.delete(cache_key))

    async def drop_many(self, ctx: ContextData, indices: List[Any] = None) -> int:
        if indices and len(indices) > 0:
            cache_keys = [self.get_cache_key(id_value) for id_value in indices]
        else:
            model_cache_key = self.get_model_cache_key()
            cache_keys = await ctx.redis.connection.keys(f'{model_cache_key}:*')

        return int(await ctx.redis.connection.delete(*cache_keys)) if cache_keys else 0

    async def drop_all(self, ctx: ContextData, prefix: Optional[str] = None) -> int:
        model_cache_key = self.get_model_cache_key()
        cache_key = f'{model_cache_key}:{prefix}*' if prefix else f'{model_cache_key}:*'
        keys_to_delete = await ctx.redis.connection.keys(cache_key)
        return int(await ctx.redis.connection.delete(*keys_to_delete)) if keys_to_delete else 0

    async def exists(self, ctx: ContextData, cache_key: str) -> bool:
        if await ctx.redis.connection.exists(cache_key):
            return await ctx.redis.connection.get(cache_key) != self._null_data
        return False

    async def scan(self, ctx: ContextData, cache_key: str = None,
                   match: str = None, count: int = None) -> AsyncIterable[Tuple[str, Optional[Dict[str, Any]]]]:
        match = f'{cache_key}:{match}' if cache_key and match else match
        async for k in ctx.redis.connection.iscan(match=match, count=count):
            cache_key = k.decode('utf-8')
            yield cache_key.split(':')[1] if ':' in cache_key else cache_key, (await self.load_one(ctx, cache_key))[1]


class HashRedisModelCache(ModelCacheProtocol):
    def __init__(self, engine: RedisStorageEngine, class_key: str,
                 serializer: Serializer, deserializer: Deserializer, ttl_enabled: bool = True):
        self._redis_engine = engine
        self._class_key = class_key
        self._serializer = serializer
        self._deserializer = deserializer
        self._ttl_enabled = ttl_enabled
        self._null_data = self._serializer.serialize(None)

    def get_model_cache_key(self) -> str:
        return f'{self._redis_engine.name}.models.{self._class_key}'

    async def cache_one(
            self, ctx: ContextData, data: ModelCacheData, ttl: Optional[timedelta] = None, **kwargs) -> bool:
        model_cache_key, cache_id = self.get_model_cache_key(), self.get_cache_id(data.cache_id)
        serialized = self._serializer.serialize(data.data)

        if self._ttl_enabled:
            res = False
            tx = ctx.redis.connection.multi_exec()
            expiration = ttl or self._redis_engine.default_ttl
            tx.hset(model_cache_key, cache_id, serialized)
            tx.expire(model_cache_key, int(expiration.total_seconds()))

            for r in await tx.execute():
                res |= r

            return bool(res)

        return bool(await ctx.redis.connection.hset(model_cache_key, cache_id, serialized))

    async def cache_many(
            self, ctx: ContextData, objects: List[ModelCacheData],
            ttl: Optional[timedelta] = None, **kwargs) -> bool:
        if not objects:
            return False

        model_cache_key = self.get_model_cache_key()
        obj_dict = {
            self.get_cache_id(instance.cache_id): self._serializer.serialize(instance.data)
            for instance in objects
        }

        if self._ttl_enabled:
            res = False
            tx = ctx.redis.connection.multi_exec()
            expiration = ttl or self._redis_engine.default_ttl
            tx.hmset_dict(model_cache_key, obj_dict)
            tx.expire(model_cache_key, int(expiration.total_seconds()))

            for r in await tx.execute():
                res |= r

            return bool(res)

        return bool(await ctx.redis.connection.hmset_dict(model_cache_key, obj_dict))

    async def load_one(
            self, ctx: ContextData, cache_key: str, **kwargs) -> Tuple[bool, Optional[Any]]:
        if ':' in cache_key:
            model_cache_key, cache_id = cache_key.split(':', maxsplit=1)
            if res := await ctx.redis.connection.hget(model_cache_key, cache_id):
                return True, self._deserializer.deserialize(res)
            return False, None
        else:
            if res := await ctx.redis.connection.get(cache_key):
                return True, self._deserializer.deserialize(res)
            return False, None

    async def load_many(
            self, ctx: ContextData, indices: List[Any], **kwargs) -> List[Optional[Dict[str, Any]]]:
        if not indices:
            return []

        cache_keys = [self.get_cache_id(id_value) for id_value in indices]
        res = await ctx.redis.connection.hmget(self.get_model_cache_key(), *cache_keys)
        objects: Dict[Any, Any] = OrderedDict()

        for id_value, data in zip(indices, res):
            objects[id_value] = self._deserializer.deserialize(data) if data is not None else None

        return list(objects.values())

    async def drop_one(self, ctx: ContextData, cache_key: str) -> bool:
        if ':' in cache_key:
            model_cache_key, cache_id = cache_key.split(':', maxsplit=1)
            return bool(await ctx.redis.connection.hdel(model_cache_key, cache_id))
        else:
            return bool(await ctx.redis.connection.delete(cache_key))

    async def drop_many(self, ctx: ContextData, indices: List[Any] = None) -> int:
        model_cache_key = self.get_model_cache_key()

        if indices and len(indices) > 0:
            cache_keys = [self.get_cache_id(id_value) for id_value in indices]
        else:
            cache_keys = await ctx.redis.connection.hkeys(model_cache_key)

        return int(await ctx.redis.connection.hdel(model_cache_key, *cache_keys)) if cache_keys else 0

    async def drop_all(self, ctx: ContextData, prefix: Optional[str] = None) -> int:
        model_cache_key = self.get_model_cache_key()
        if not prefix:
            keys_to_delete = await ctx.redis.connection.hkeys(model_cache_key)
        else:
            keys_to_delete = list()
            for cache_key in await ctx.redis.connection.hkeys(model_cache_key, encoding='ascii'):
                if cache_key.startswith(prefix):
                    keys_to_delete.append(cache_key)

        return int(await ctx.redis.connection.hdel(model_cache_key, *keys_to_delete)) if keys_to_delete else 0

    async def exists(self, ctx: ContextData, cache_key: str) -> bool:
        if ':' in cache_key:
            model_cache_key, cache_id = cache_key.split(':', maxsplit=1)
            e = bool(await ctx.redis.connection.hexists(model_cache_key, cache_id))
        else:
            e = bool(await ctx.redis.connection.exists(cache_key))
        if e:
            return await ctx.redis.connection.get(cache_key) != self._null_data
        return False

    async def scan(self, ctx: ContextData, cache_key: str = None,
                   match: str = None, count: int = None) -> AsyncIterable[Tuple[str, Optional[Dict[str, Any]]]]:
        async for k, data in ctx.redis.connection.ihscan(cache_key, match=match, count=count):
            yield k.decode('utf-8'), self._deserializer.deserialize(data)



from typing import Optional

from fastapi import Cookie, Query

from dropland.data.context import get_context


async def query_locale(locale: Optional[str] = Query(None)) -> Optional[str]:
    if locale:
        get_context().data['lang'] = locale
    return locale


async def cookie_locale(locale: Optional[str] = Cookie(None)) -> Optional[str]:
    if locale:
        get_context().data['lang'] = locale
    return locale



from pathlib import Path
from typing import Any, Dict, Optional

from pydantic.class_validators import validator
from pydantic.env_settings import BaseSettings


class RmqSettings(BaseSettings):
    RMQ_HOST: str = 'localhost'
    RMQ_PORT: int = 5672
    RMQ_USER: str = 'guest'
    RMQ_PASSWORD: str = 'guest'
    RMQ_PASSWORD_FILE: Optional[Path] = None
    RMQ_VIRTUALHOST: str = '/'
    RMQ_TIMEOUT_SECONDS: int = 5
    RMQ_POOL_MAX_CONNECTIONS: int = 4
    RMQ_POOL_MAX_CHANNELS_PER_CONNECTION: int = 100

    RMQ_URI: Optional[str] = None

    @validator('RMQ_URI', pre=True)
    def assemble_rmq_connection(cls, v: Optional[str], values: Dict[str, Any]) -> str:
        if isinstance(v, str):
            return v

        if 'RMQ_PASSWORD_FILE' in values:
            password_file = values.get('RMQ_PASSWORD_FILE')
            if isinstance(password_file, Path) and password_file.exists():
                with password_file.open() as f:
                    values['RMQ_PASSWORD'] = f.read().strip()

        return f'amqp://{values.get("RMQ_USER")}:{values.get("RMQ_PASSWORD")}' \
               f'@{values.get("RMQ_HOST")}:{values.get("RMQ_PORT")}/'

    class Config:
        case_sensitive = False
        env_file = '.env'



from abc import ABC, abstractmethod
from inspect import isclass
from typing import List, Callable, Set, Tuple, Optional, Union, Type


class BasePermission(ABC):
    depends_on = []

    @classmethod
    @abstractmethod
    def has_permissions(cls, *args, **kwargs) -> bool:
        raise NotImplementedError

    # noinspection PyMethodMayBeStatic
    def has_object_permissions(self, *args, **kwargs) -> bool:
        return False


PermissionType = Union[Type, object, Callable[..., bool]]


class PermissionChecker:
    def __init__(self, permissions: List[PermissionType]):
        self.permissions = permissions

    def _check_permission(self, permission: PermissionType,
                          perms_cache: Set[str], *args, **kwargs) -> Tuple[bool, Optional[PermissionType]]:
        perm_name = None

        if isclass(permission):
            curr_deps = list(getattr(permission, 'depends_on', []))
            perm_name = '.'.join([permission.__module__, permission.__qualname__])

            if perm_name in perms_cache:
                return True, None

            for dep_permission in curr_deps:
                if not self._check_permission(dep_permission, perms_cache, *args, **kwargs):
                    return False, dep_permission

        elif hasattr(permission, '__name__'):
            perm_name = permission.__name__

        if isclass(permission):
            if issubclass(permission, BasePermission):
                if not permission.has_permissions(*args, **kwargs):
                    return False, permission
        elif isinstance(permission, BasePermission):
            if not permission.has_object_permissions(*args, **kwargs):
                return False, permission
        elif not permission(*args, **kwargs):
            return False, permission

        if perm_name:
            perms_cache.add(perm_name)

        return True, None

    def __call__(self, *args, **kwargs) -> Tuple[bool, Optional[PermissionType]]:
        perms_cache: Set[str] = set()

        for permission in self.permissions:
            res, error_callable = self._check_permission(permission, perms_cache, *args, **kwargs)
            if not res:
                return False, error_callable
        return True, None



from dependency_injector import containers, providers

from app.module import Module

from .redis import USE_REDIS
from .rmq import USE_RMQ
from .scheduler import USE_SCHEDULER
from .sql import USE_SQL
from dependency_injector import containers, providers

from app.module import Module
from .redis import USE_REDIS
from .rmq import USE_RMQ
from .scheduler import USE_SCHEDULER
from .sql import USE_SQL

__blocks__ = dict()


if USE_REDIS:
    from .redis.containers import RedisBlock
    __blocks__['redis'] = RedisBlock

if USE_RMQ:
    from .rmq.containers import RmqBlock
    __blocks__['rmq'] = RmqBlock

if USE_SCHEDULER:
    from .scheduler.containers import SchedulerBlock
    __blocks__['scheduler'] = SchedulerBlock

if USE_SQL:
    from .sql.containers import SqlBlock
    __blocks__['sql'] = SqlBlock


class BlocksContainer(containers.DeclarativeContainer):
    __self__ = providers.Self()

    blocks = providers.Aggregate(**{
        name: providers.Factory(class_) for name, class_ in __blocks__.items()
    })


class BlocksModule(Module):
    pass




try:
    import aioredis

    from .engine import EngineConfig, RmqStorageBackend, RmqStorageEngine
    from .session import SessionManager
    from .settings import RmqSettings

    USE_RMQ = True

except ImportError:
    USE_RMQ = False



from collections import defaultdict
from dataclasses import dataclass, field
from types import ClassMethodDescriptorType, FunctionType, GetSetDescriptorType, MappingProxyType, MemberDescriptorType, \
    MethodDescriptorType, MethodType, MethodWrapperType, WrapperDescriptorType
from typing import Protocol, Set, Dict, List, Any, Optional

from pydantic.main import BaseModel as PydanticModel

from .base import ModelProtocol
from .relationship import Relationship
from ..context import ContextData


@dataclass
class FieldsCache:
    priv: Set[str] = field(default_factory=set)
    pub: Set[str] = field(default_factory=set)
    ser: Set[str] = field(default_factory=set)


class ModelFieldsMixin(Protocol):
    class Meta:
        private_fields: Set[str] = set()
        public_fields: Set[str] = set()
        serializable_fields: Set[str] = set()
        non_serializable_fields: Set[str] = set()
        relationships: Dict[str, Relationship] = dict()
        _fields_cache: Dict[str, FieldsCache] = dict()

    @classmethod
    def _fields_cache_key(cls):
        return '.'.join([cls.__module__, cls.__qualname__])

    # noinspection PyProtectedMember
    @classmethod
    def _calculate_fields(cls):
        private_types = (
            type, FunctionType, MethodType, MappingProxyType,
            WrapperDescriptorType, MethodWrapperType, MethodDescriptorType,
            ClassMethodDescriptorType, GetSetDescriptorType, MemberDescriptorType
        )

        private_fields, public_fields, serializable_fields = set(), set(), set()
        relationship_keys = cls.Meta.relationships.keys()

        for attr in dir(cls):
            value = getattr(cls, attr)
            if isinstance(value, private_types) or attr[0] == '_':
                private_fields.add(attr)
            elif not isinstance(value, private_types):
                if not isinstance(value, property) and attr not in relationship_keys:
                    serializable_fields.add(attr)
                public_fields.add(attr)

        private_fields.update(cls.Meta.private_fields)
        private_fields.difference_update(cls.Meta.public_fields)
        public_fields.update(cls.Meta.public_fields)
        public_fields.difference_update(cls.Meta.private_fields)
        serializable_fields.update(cls.Meta.serializable_fields)
        serializable_fields.difference_update(cls.Meta.non_serializable_fields)
        cls.Meta._fields_cache[cls._fields_cache_key()] = \
            FieldsCache(priv=private_fields, pub=public_fields, ser=serializable_fields)

    # noinspection PyProtectedMember
    @classmethod
    def drop_fields_cache(cls):
        key = cls._fields_cache_key()
        cls.Meta._fields_cache.pop(key, None)

    # noinspection PyProtectedMember
    @classmethod
    def get_private_fields(cls) -> Set[str]:
        key = cls._fields_cache_key()
        if key not in cls.Meta._fields_cache:
            cls._calculate_fields()
        return cls.Meta._fields_cache[key].priv

    # noinspection PyProtectedMember
    @classmethod
    def get_public_fields(cls) -> Set[str]:
        key = cls._fields_cache_key()
        if key not in cls.Meta._fields_cache:
            cls._calculate_fields()
        return cls.Meta._fields_cache[key].pub

    @classmethod
    def get_fields(cls) -> Set[str]:
        return cls.get_public_fields()

    def get_values(
        self, only_fields: List[str] = None,
            exclude_fields: List[str] = None) -> Dict[str, Any]:
        public_fields = self.get_public_fields()
        only_fields = set(only_fields) if only_fields else set()
        exclude_fields = set(exclude_fields) if exclude_fields else set()

        return {
            name: getattr(self, name) for name in public_fields
            if hasattr(self, name) and (not only_fields or name in only_fields)
            and (not exclude_fields or name not in exclude_fields)
        }

    def _assign(self, data: Dict[str, Any]) -> 'ModelFieldsMixin':
        for k, v in data.items():
            setattr(self, k, v)
        return self

    @classmethod
    def _construct(cls, data: Dict[str, Any], instance: Optional['ModelFieldsMixin'] = None,
                   only_fields: Optional[Set[str]] = None, **kwargs) -> Optional['ModelFieldsMixin']:
        if not isinstance(data, dict):
            return None

        field_data = dict() if only_fields else data
        if only_fields:
            for k, v in data.items():
                if only_fields is None or k in only_fields:
                    field_data[k] = v

        if instance:
            return instance._assign(field_data)
        else:
            if issubclass(cls, PydanticModel):
                # noinspection PyArgumentList
                return cls(**field_data)
            else:
                return cls()._assign(field_data)

    @classmethod
    def _construct_list(cls, objects: List[Dict[str, Any]], **kwargs) -> List['ModelFieldsMixin']:
        return [cls._construct(data, **kwargs) for data in objects if data is not None]

    @classmethod
    async def _register_instances(cls, ctx: ContextData, objects: List[ModelProtocol]):
        pass

    @classmethod
    async def _unregister_indices(cls, ctx: ContextData, indices: List[Any] = None):
        pass

    @classmethod
    async def _build_rela(
        cls, ctx: ContextData, instance: ModelProtocol,
            load_fields: Optional[Set[str]] = None, **kwargs) -> Optional[ModelProtocol]:
        res = await cls._build_relationships(ctx, [instance], load_fields, **kwargs)
        return res[0]

    @classmethod
    async def _build_rela_list(
        cls, ctx: ContextData, objects: List[ModelProtocol],
            load_fields: Optional[Set[str]] = None, **kwargs) -> List[ModelProtocol]:
        return await cls._build_relationships(ctx, objects, load_fields, **kwargs)

    @classmethod
    async def _build_relationships(
        cls, _: ContextData, objects: List[ModelProtocol],
            load_fields: Optional[Set[str]] = None, **kwargs) -> List[ModelProtocol]:
        deps_ids: Dict[str, Dict[Any, Set[Any]]] = defaultdict(lambda: defaultdict(set))
        deps_values: Dict[str, Dict[Any, Any]] = defaultdict(dict)
        rela_map = cls.Meta.relationships.items()
        load_all = load_fields is None

        for instance in objects:
            if instance is None:
                continue

            for dep_key, relationship in rela_map:
                if not load_all and dep_key not in load_fields:
                    continue

                dep_key_value = relationship.get_key(instance)

                if relationship.single:
                    dep_key_value = [dep_key_value]

                deps_ids[dep_key][instance.get_id_value()].update(set(dep_key_value))

        for dep_key, relationship in rela_map:
            if not load_all and dep_key not in load_fields:
                continue

            dep_keys = set()
            for v in deps_ids[dep_key].values():
                dep_keys.update(v)
            model_class = relationship.get_model()
            dep_values = await model_class.get_any(list(dep_keys), **kwargs)

            for k, v in zip(dep_keys, dep_values):
                deps_values[dep_key][k] = v

        for instance in objects:
            if instance is None:
                continue

            for dep_key, relationship in rela_map:
                if not load_all and dep_key not in load_fields:
                    continue

                dep_key_values = list(deps_ids[dep_key][instance.get_id_value()])

                if relationship.single:
                    dep_key_value = dep_key_values[0] if len(dep_key_values) > 0 else dep_key_values
                    dep_value = deps_values[dep_key][dep_key_value]
                    setattr(instance, dep_key, dep_value)
                else:
                    dep_values = [deps_values[dep_key][k] for k in dep_key_values]
                    setattr(instance, dep_key, dep_values)

        return objects



from collections import defaultdict
from typing import Dict, List

from dropland.blocks.sql.base import SqlStorageType
from dropland.core.loaders.backend import BackendLoader
from .engine import SqlStorageBackend


class SqlBackendLoader(BackendLoader[SqlStorageBackend]):
    def __init__(self):
        super().__init__('dropland.blocks.sql', 'create_sql_storage_backend')
        self._db_support_dict: Dict[SqlStorageType, List[str]] = defaultdict(list)

    def on_load(self, backend: SqlStorageBackend):
        for db_type in backend.db_supports:
            self._db_support_dict[db_type].append(backend.name)

    def get_backends_for_db(self, db_type: SqlStorageType) -> List[str]:
        return self._db_support_dict.get(db_type, [])



from datetime import datetime

import pytest

from dropland.blocks.redis import USE_REDIS

pytestmark = pytest.mark.skipif(not USE_REDIS, reason='For Redis only')

if USE_REDIS:
    from dropland.blocks.redis.model import RedisCacheType, RedisMethodCache
    from .redis_models_data import redis_engine


@pytest.mark.asyncio
@pytest.mark.parametrize('cache_type', (RedisCacheType.SIMPLE, RedisCacheType.HASH), ids=['simple', 'hash'])
async def test_method_cache(cache_type, redis_session):
    method_cache = RedisMethodCache(redis_engine, 'test', cache_type)

    assert method_cache.get_model_cache_key() == 'dropland.models.test'
    assert method_cache.get_cache_id(123) == '123'
    assert method_cache.get_cache_key(1) == 'dropland.models.test:1'

    args = {'a': 1, 'b': 12.34, 'c': 'asdf', 'd': datetime.now()}
    exists, data = await method_cache.get('method', args)
    assert not exists
    assert not data

    assert await method_cache.put('method', None, None)
    exists, data = await method_cache.get('method', args)
    assert not exists
    assert not data
    exists, data = await method_cache.get('method', None)
    assert exists
    assert data is None

    assert await method_cache.put('method', args, None)
    exists, data = await method_cache.get('method', args)
    assert exists
    assert data is None
    exists, data = await method_cache.get('method', None)
    assert exists
    assert data is None

    assert await method_cache.put('method', args, 123)
    exists, data = await method_cache.get('method', args)
    assert exists
    assert data is 123
    exists, data = await method_cache.get('method', None)
    assert exists
    assert data is None

    assert await method_cache.put('method', None, '456')
    exists, data = await method_cache.get('method', args)
    assert exists
    assert data is 123
    exists, data = await method_cache.get('method', None)
    assert exists
    assert data == '456'

    assert await method_cache.drop('method', args)
    assert not await method_cache.drop('method', args)
    exists, data = await method_cache.get('method', args)
    assert not exists
    assert not data
    exists, data = await method_cache.get('method', None)
    assert exists
    assert data == '456'

    assert await method_cache.drop('method', None)
    assert not await method_cache.drop('method', None)
    exists, data = await method_cache.get('method', args)
    assert not exists
    assert not data
    exists, data = await method_cache.get('method', None)
    assert not exists
    assert not data



from typing import Dict, List, Optional

from dependency_injector.containers import Container

from . import base


class Module(base.Module, base.ContainerBlock):
    def __init__(self, container: Container, name: str):
        super().__init__(container)

        self._name = name
        self._resource_containers: Dict[str, Container] = dict()
        self._resources: Dict[str, base.ResourceBlock] = dict()

        if hasattr(container, 'blocks'):
            for name, provider in container.blocks.providers.items():
                block_container = container.blocks(name)
                self._resource_containers[name] = block_container
                self._resources[name] = block_container.block(block_container)

    @property
    def name(self) -> str:
        return self._name

    @property
    def resources(self) -> List[base.ResourceBlock]:
        return list(self._resources.values())

    def get_resource(self, name) -> Optional[base.ResourceBlock]:
        return self._resources.get(name)

    def sync_startup(self, application=None, service=None, *args, **kwargs):
        for r in self.resources:
            r.sync_startup(application=application, service=service, module=self, *args, **kwargs)

    def sync_shutdown(self, application=None, service=None, *args, **kwargs):
        for r in self.resources:
            r.sync_shutdown(application=application, service=service, module=self, *args, **kwargs)

    async def startup(self, application=None, service=None, *args, **kwargs):
        for r in self.resources:
            await r.startup(application=application, service=service, module=self, *args, **kwargs)

    async def shutdown(self, application=None, service=None, *args, **kwargs):
        for r in self.resources:
            await r.shutdown(application=application, service=service, module=self, *args, **kwargs)


class SessionModule(Module, base.SessionResourceBlock):
    def sync_session_begin(self, application=None, service=None, module=None, *args, **kwargs):
        for r in self.resources:
            if not isinstance(r, base.SessionResourceBlock):
                continue
            r.sync_session_begin(application=application, service=service, module=self, *args, **kwargs)

    def sync_session_finish(self, application=None, service=None, module=None, *args, **kwargs):
        for r in self.resources:
            if not isinstance(r, base.SessionResourceBlock):
                continue
            r.sync_session_finish(application=application, service=service, module=self, *args, **kwargs)

    async def session_begin(self, application=None, service=None, module=None, *args, **kwargs):
        for r in self.resources:
            if not isinstance(r, base.SessionResourceBlock):
                continue
            await r.session_begin(application=application, service=service, module=self, *args, **kwargs)

    async def session_finish(self, application=None, service=None, module=None, *args, **kwargs):
        for r in self.resources:
            if not isinstance(r, base.SessionResourceBlock):
                continue
            await r.session_finish(application=application, service=service, module=self, *args, **kwargs)



import contextlib
from pathlib import Path
from typing import Dict, List, Optional

from dependency_injector import providers
from dependency_injector.containers import Container
from dependency_injector.wiring import Provide, inject

from dropland.data.context import with_context
from dropland.log import logger
from dropland.util import import_path, invoke_async, invoke_sync
from . import base, service


class Application(base.Application, service.SessionService):
    def __init__(self, container: Container, name: str, debug: bool = False, version: str = ''):
        super().__init__(container, name)

        self._version = version
        self._debug = debug
        self._path = self._inspect_cwd()

        self._imported_modules = None
        self._service_containers: Dict[str, Container] = dict()
        self._services: Dict[str, service.Service] = dict()

        if hasattr(container, 'services'):
            for name, provider in container.services.providers.items():
                service_container = container.services(name)
                self._service_containers[name] = service_container
                self._services[name] = service_container.instance(service_container, name)

        container.instance = providers.Object(self)
        container.wire(modules=[__name__])

    @property
    def services(self) -> List[base.Service]:
        return list(self._services.values())

    def get_service(self, name) -> Optional[base.Service]:
        return self._services.get(name)

    @staticmethod
    def _inspect_cwd():
        import inspect
        return Path(inspect.stack()[3].filename).parent.absolute()

    @property
    def version(self) -> str:
        return self._version

    @property
    def debug(self) -> bool:
        return self._debug

    def get_cwd(self) -> Path:
        return self._path.absolute()

    def sync_startup(self, *args, **kwargs):
        super().sync_startup(application=self, *args, **kwargs)

        for s in self.services:
            s.sync_startup(application=self, *args, **kwargs)

    def sync_shutdown(self, *args, **kwargs):
        for s in self.services:
            s.sync_shutdown(application=self, *args, **kwargs)

        super().sync_shutdown(application=self, *args, **kwargs)

    async def startup(self, *args, **kwargs):
        await super().startup(application=self, *args, **kwargs)

        for s in self.services:
            await s.startup(application=self, *args, **kwargs)

    async def shutdown(self, *args, **kwargs):
        for s in self.services:
            await s.shutdown(application=self, *args, **kwargs)

        await super().shutdown(application=self, *args, **kwargs)

    def sync_session_begin(self, application=None, service=None, module=None, *args, **kwargs):
        super().sync_session_begin(application=self, *args, **kwargs)

        for s in self.services:
            if not isinstance(s, base.SessionResourceBlock):
                continue
            s.sync_session_begin(application=self, *args, **kwargs)

    def sync_session_finish(self, application=None, service=None, module=None, *args, **kwargs):
        for s in self.services:
            if not isinstance(s, base.SessionResourceBlock):
                continue
            s.sync_session_finish(application=self, *args, **kwargs)

        super().sync_session_finish(application=self, *args, **kwargs)

    async def session_begin(self, application=None, service=None, module=None, *args, **kwargs):
        await super().session_begin(application=self, *args, **kwargs)

        for s in self.services:
            if not isinstance(s, base.SessionResourceBlock):
                continue
            await s.session_begin(application=self, *args, **kwargs)

    async def session_finish(self, application=None, service=None, module=None, *args, **kwargs):
        for s in self.services:
            if not isinstance(s, base.SessionResourceBlock):
                continue
            await s.session_finish(application=self, *args, **kwargs)

        await super().session_finish(application=self, *args, **kwargs)

    def _import_modules(self, app_name: Optional[str] = None):
        from tomlkit import parse

        modules, result = dict(), dict()
        pyproject_toml_path = self.get_cwd() / 'pyproject.toml'

        if pyproject_toml_path.exists():
            with open(str(pyproject_toml_path)) as f:
                pyproject_toml = parse(string=f.read())

            if project := pyproject_toml.get('project'):
                if app_name:
                    if section := project.get(app_name):
                        modules = section.get('modules', dict())
                else:
                    modules = project.get('modules', dict())

        app_name = app_name or 'module'

        for name, dotted_path in modules.items():
            logger.info(f'Import module: {name}')
            dotted_path = dotted_path if ':' in dotted_path else f'{dotted_path}:{app_name}_init'

            if entrypoint := import_path(dotted_path):
                result[name] = entrypoint

        return result

    def sync_load_modules(self, *args, **kwargs):
        if not self._imported_modules:
            self._imported_modules = self._import_modules(self.name)

        with self.sync_with_app_sessions():
            for name, entrypoint in self._imported_modules.items():
                if name not in self._modules:
                    logger.warn(f'Application "{self.name}" does not contains module "{name}"')
                    continue

                logger.info(f'Application "{self.name}": load module "{name}"')

                module_args = (self, self._modules[name], *args)
                entrypoint(*module_args, **kwargs)

    async def load_modules(self, *args, **kwargs):
        if not self._imported_modules:
            self._imported_modules = self._import_modules(self.name)

        async with self.with_app_sessions():
            for name, entrypoint in self._imported_modules.items():
                if name not in self._modules:
                    logger.warn(f'Application "{self.name}" does not contains module "{name}"')
                    continue

                logger.info(f'Application "{self.name}": load module "{name}"')

                module_args = (self, self._modules[name], *args)
                await entrypoint(*module_args, **kwargs)

    @contextlib.contextmanager
    def sync_with_app_resources(self):
        with with_context(False):
            with contextlib.ExitStack() as stack:
                invoke_sync(self.startup)
                self.sync_startup()
                stack.callback(self.sync_shutdown)
                stack.callback(invoke_sync, self.shutdown)
                yield

    @contextlib.asynccontextmanager
    async def with_app_resources(self):
        with with_context(False):
            async with contextlib.AsyncExitStack() as stack:
                await self.startup()
                await invoke_async(self.sync_startup)
                stack.push_async_callback(invoke_async, self.sync_shutdown)
                stack.push_async_callback(self.shutdown)
                yield

    @contextlib.contextmanager
    def sync_with_app_sessions(self):
        with with_context(False):
            with contextlib.ExitStack() as stack:
                invoke_sync(self.session_begin)
                self.sync_session_begin()
                stack.callback(self.sync_session_finish)
                stack.callback(invoke_sync, self.session_finish)
                yield

    @contextlib.asynccontextmanager
    async def with_app_sessions(self):
        with with_context(False):
            async with contextlib.AsyncExitStack() as stack:
                await self.session_begin()
                await invoke_async(self.sync_session_begin)
                stack.push_async_callback(invoke_async, self.sync_session_finish)
                stack.push_async_callback(self.session_finish)
                yield


@inject
def current_application(_app: Application = Provide['instance']):
    return _app



import asyncio

import pytest

from dropland.blocks.redis import USE_REDIS
from dropland.blocks.rmq import USE_RMQ
from dropland.blocks.sql import USE_SQL
from dropland.blocks.sql.base import SqlStorageType

if USE_REDIS:
    from dropland.blocks.redis.containers import RedisStorage, RedisBlock
    from dropland.blocks.redis.engine import EngineConfig as RedisEngineConfig

if USE_RMQ:
    from dropland.blocks.rmq.containers import RmqStorage, RmqBlock
    from dropland.blocks.rmq.engine import EngineConfig as RmqEngineConfig

if USE_SQL:
    from dropland.blocks.sql.containers import MultipleSqlStorage, SqlStorage, SqlBlock
    from dropland.blocks.sql import EngineConfig as SqlEngineConfig

from tests import MYSQL_URI, POSTGRES_URI, REDIS_URI, RMQ_URI, SQLITE_URI


if USE_SQL:
    sql_storage = MultipleSqlStorage()

    sql_storage.config.from_dict({
        'sqlite': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': SqlEngineConfig(
                url=SQLITE_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': False
        },
        'asqlite': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': SqlEngineConfig(
                url=SQLITE_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': True
        },
        'pg': {
            'db_type': SqlStorageType.POSTGRES,
            'engine_config': SqlEngineConfig(
                url=POSTGRES_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': False
        },
        'apg': {
            'db_type': SqlStorageType.POSTGRES,
            'engine_config': SqlEngineConfig(
                url=POSTGRES_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': True
        },
        'ms': {
            'db_type': SqlStorageType.MYSQL,
            'engine_config': SqlEngineConfig(
                url=MYSQL_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': False
        },
        'ams': {
            'db_type': SqlStorageType.MYSQL,
            'engine_config': SqlEngineConfig(
                url=MYSQL_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': True
        },
    })


@pytest.fixture(scope='session')
def test_sql_storage():
    if USE_SQL:
        return SqlStorage()
    else:
        return None


@pytest.fixture(scope='session')
def test_redis_storage():
    if USE_REDIS:
        return RedisStorage()
    else:
        return None


@pytest.fixture(scope='session')
def test_rmq_storage():
    if USE_RMQ:
        return RmqStorage()
    else:
        return None


@pytest.fixture(scope='session')
def sql_sync_engine(test_sql_storage):
    conf = sql_storage.config.get('sqlite')
    return test_sql_storage.create_engine(
        'dropland-test-sync', conf['engine_config'], conf['db_type'], conf.get('use_async')
    )


@pytest.fixture(scope='session')
def sql_async_engine(test_sql_storage):
    conf = sql_storage.config.get('asqlite')
    return test_sql_storage.create_engine(
        'dropland-test-async', conf['engine_config'], conf['db_type'], conf.get('use_async')
    )


@pytest.fixture(scope='session')
def sqlite_engine():
    return sql_storage.create_engine('sqlite')


@pytest.fixture(scope='session')
def sqlite_async_engine():
    return sql_storage.create_engine('asqlite')


@pytest.fixture(scope='session')
def pg_engine():
    return sql_storage.create_engine('pg')


@pytest.fixture(scope='session')
def pg_async_engine():
    return sql_storage.create_engine('apg')


@pytest.fixture(scope='session')
def mysql_engine():
    return sql_storage.create_engine('ms')


@pytest.fixture(scope='session')
def mysql_async_engine():
    return sql_storage.create_engine('ams')


@pytest.fixture(scope='session')
def redis_engine(test_redis_storage):
    engine_config = RedisEngineConfig(url=REDIS_URI)
    return test_redis_storage.create_engine('dropland', engine_config)


@pytest.fixture(scope='session')
def rmq_engine(test_rmq_storage):
    engine_config = RmqEngineConfig(url=RMQ_URI)
    return test_rmq_storage.create_engine('dropland', engine_config)


@pytest.fixture(scope='session')
def event_loop():
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()


if USE_SQL:
    @pytest.fixture(scope='function')
    def sync_sql_session(event_loop, test_sql_storage):
        block = SqlBlock(test_sql_storage)
        block.sync_startup()
        block.sync_session_begin(autocommit=False)
        yield test_sql_storage.manager()
        block.sync_session_finish()
        block.sync_shutdown()

    @pytest.fixture(scope='function')
    async def async_sql_session(event_loop, test_sql_storage):
        block = SqlBlock(test_sql_storage)
        await block.startup()
        await block.session_begin(autocommit=False)
        yield test_sql_storage.manager()
        await block.session_finish()
        await block.shutdown()


if USE_REDIS:
    @pytest.fixture(scope='module')
    def wire_redis_storage(test_redis_storage):
        test_redis_storage.wire()
        yield
        test_redis_storage.unwire()

    @pytest.fixture(scope='function')
    async def redis_session(event_loop, wire_redis_storage, test_redis_storage, redis_engine):
        block = RedisBlock(test_redis_storage)
        await block.startup()
        await block.session_begin()
        yield test_redis_storage.manager()
        await block.session_finish()
        await block.shutdown()


if USE_RMQ:
    @pytest.fixture(scope='module')
    def wire_rmq_storage(test_rmq_storage):
        test_rmq_storage.wire()
        yield
        test_rmq_storage.unwire()

    @pytest.fixture(scope='function')
    async def rmq_session(event_loop, wire_rmq_storage, test_rmq_storage, rmq_engine):
        block = RmqBlock(test_rmq_storage)
        await block.startup()
        await block.session_begin()
        yield test_rmq_storage.manager()
        await block.session_finish()
        await block.shutdown()



from abc import ABC
from typing import List

from fastapi import HTTPException, params
from starlette import status
from starlette.requests import Request

from dropland.permissions import BasePermission, PermissionChecker, PermissionType


class HttpPermission(BasePermission, ABC):
    error_msg = 'Forbidden'
    status_code = status.HTTP_403_FORBIDDEN


class PermDepends(params.Depends, PermissionChecker):
    def __init__(self, permissions: List[PermissionType]):
        params.Depends.__init__(self, dependency=self.__call__)
        PermissionChecker.__init__(self, permissions)

    def __call__(self, request: Request):
        res, error_callable = super().__call__(request)
        if not res:
            if isinstance(error_callable, HttpPermission):
                raise HTTPException(status_code=error_callable.status_code, detail=error_callable.error_msg)
            else:
                raise HTTPException(status_code=HttpPermission.status_code, detail=HttpPermission.error_msg)


class Authenticated(HttpPermission):
    error_msg = 'You must be authenticated'

    @classmethod
    def has_permissions(cls, request: Request):
        return request.user.is_authenticated


class Admin(HttpPermission):
    error_msg = 'Admin area'
    depends_on = [Authenticated]

    @classmethod
    def has_permissions(cls, request: Request):
        return bool(request.user.admin)



import pytest

from dropland.blocks.rmq import USE_RMQ

pytestmark = pytest.mark.skipif(not USE_RMQ, reason='For RabbitMQ only')

if USE_RMQ:
    from dropland.blocks.rmq.containers import RmqStorage, SingleRmqStorage, MultipleRmqStorage, RmqBlock
    from dropland.blocks.rmq.engine import EngineConfig
    from tests import RMQ_URI


def test_session_connections(test_rmq_storage, rmq_engine, rmq_session):
    assert not rmq_session.get_session('_')

    session = rmq_session.get_session('dropland')
    assert session
    assert session.engine is rmq_engine
    assert session.connection

    created, session = rmq_session.get_or_create_session('dropland')
    assert not created
    assert session
    assert session.engine is rmq_engine
    assert session.connection

    assert not test_rmq_storage.get_session('_')
    assert test_rmq_storage.get_session('dropland') is session


@pytest.mark.asyncio
async def test_container_sessions():
    config = EngineConfig(url=RMQ_URI)

    cont = RmqStorage()
    engine = cont.create_engine('dropland', config)
    block = RmqBlock(cont)
    await block.startup()
    await block.session_begin()

    session = cont.get_session('dropland')
    assert session
    assert session.engine is engine
    assert session.connection
    assert not cont.get_session('_')
    await block.session_finish()
    await block.shutdown()
    cont.unwire()

    cont = SingleRmqStorage()
    cont.config.from_dict({
        'name': 'dropland1',
        'engine_config': config
    })
    engine = cont.create_engine()
    block = RmqBlock(cont)
    await block.startup()
    await block.session_begin()

    session = cont.get_session()
    assert session
    assert session.engine is engine
    assert session.connection
    await block.session_finish()
    await block.shutdown()
    cont.unwire()

    cont = MultipleRmqStorage()
    cont.config.from_dict({
        'one': {
            'name': 'dropland1',
            'engine_config': config
        },
        'two': {
            'name': 'dropland2',
            'engine_config': config
        },
    })
    engine1 = cont.create_engine('one')
    engine2 = cont.create_engine('two')
    block = RmqBlock(cont)
    await block.startup()
    await block.session_begin()

    for name, engine in zip(('one', 'two'), (engine1, engine2)):
        session = cont.get_session(name)
        assert session
        assert session.engine is engine
        assert session.connection

    assert not cont.get_session('_')
    await block.session_finish()
    await block.shutdown()
    cont.unwire()



import asyncio
import contextlib
import itertools
from functools import partial
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from dependency_injector.wiring import Provide, inject
from sqlalchemy import Column, delete, distinct, exists, func, inspect, literal_column, select, tuple_, update
from sqlalchemy.engine import Row
from sqlalchemy.exc import NoInspectionAvailable
from sqlalchemy.orm import DeclarativeMeta, RelationshipProperty, attributes, declarative_base, joinedload, registry, \
    selectinload
from sqlalchemy.orm.instrumentation import instance_state
from sqlalchemy.orm.util import identity_key
from sqlalchemy.sql import ColumnCollection

from dropland.data.context import ContextData, get_context
from dropland.data.models.sql import CreateSchemaType, SqlModel, SqlTableInfo, UpdateSchemaType
from dropland.util import calculate_digest
from .base import SqlStorageType
from .engine import SqlStorageEngine


class SqlaModel(DeclarativeMeta, type(SqlModel)):
    # noinspection PyUnresolvedReferences
    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            column_names_map = inspect(cls).c
        except NoInspectionAvailable:
            pass
        else:
            column_names_map = zip([column.name for column in column_names_map], column_names_map.keys())
            cls.Meta._column_names_map = dict(column_names_map)


SqlaModelMeta = declarative_base(name='SqlaModel', cls=(SqlModel,), metaclass=SqlaModel)


class SqlaModelBase(SqlaModelMeta):
    __abstract__ = True

    class Meta(SqlModel.Meta):
        single_loader = joinedload
        list_loader = selectinload
        _sql_engine = None
        _column_names_map = dict()
        _query_opts_cache: Dict[str, list] = dict()

    def __init_subclass__(cls, sql_engine: SqlStorageEngine = None, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.Meta.private_fields.update({'metadata', 'registry'})
        cls.Meta.non_serializable_fields.update({'metadata', 'registry'})
        cls.Meta._query_opts_cache = dict()

        if sql_engine:
            cls.Meta._sql_engine = sql_engine

            if not hasattr(cls, 'registry'):
                cls.registry = registry()

            cls.metadata = sql_engine.metadata

    def get_id_value(self) -> Any:
        return self._get_column_values(self._get_id_columns())

    # noinspection PyProtectedMember
    @classmethod
    def get_engine(cls) -> SqlStorageEngine:
        return cls.Meta._sql_engine

    @classmethod
    def get_table_info(cls) -> SqlTableInfo:
        engine = cls.get_engine()
        return SqlTableInfo(db_type=engine.db_type, is_async=engine.is_async)

    @classmethod
    @contextlib.contextmanager
    @inject
    def _session_context(cls, sql_block=Provide['block']):
        info, ctx = cls.get_table_info(), get_context()
        assert not info.is_async, 'Use _async_session_context() function instead'

        if 'sql' in ctx and ctx.sql is not None:
            yield ctx
        else:
            ctx.sql = sql_block.get_session(cls.get_engine().name)
            yield ctx
            if 'sql' in ctx:
                del ctx.sql

    @classmethod
    @contextlib.asynccontextmanager
    @inject
    async def _async_session_context(cls, sql_block=Provide['block']):
        info, ctx = cls.get_table_info(), get_context()
        assert info.is_async, 'Use _session_context() function instead'

        if '_sql_cond' not in ctx:
            ctx._sql_cnt = 0
            ctx._sql_cond = asyncio.Condition()

        async with ctx._sql_cond:
            if 'sql' in ctx:
                ctx._sql_cnt += 1
                ctx._sql_cond.release()
                try:
                    yield ctx
                finally:
                    await ctx._sql_cond.acquire()
                    ctx._sql_cnt -= 1
                    if 0 == ctx._sql_cnt:
                        ctx._sql_cond.notify()
            else:
                engine_name = cls.get_engine().name
                async with sql_block.container.manager().session_context(engine_name):
                    ctx.sql = sql_block.get_session(engine_name)
                    assert ctx.sql
                    ctx._sql_cnt = 1
                    ctx._sql_cond.release()
                    try:
                        yield ctx
                    finally:
                        await ctx._sql_cond.acquire()
                        ctx._sql_cnt -= 1
                        if 0 == ctx._sql_cnt:
                            del ctx.sql
                        else:
                            await ctx._sql_cond.wait_for(lambda: 0 == ctx._sql_cnt)
                            del ctx.sql

    @classmethod
    def get_columns(cls) -> ColumnCollection:
        return cls.__table__.columns

    @classmethod
    def get_relationships(cls) -> Dict[str, Any]:
        return cls.__mapper__.relationships.items()

    # noinspection PyProtectedMember
    @classmethod
    def query_options(cls, include: Optional[Set[str]] = None, exclude: Optional[Set[str]] = None):
        args_digest = calculate_digest((include, exclude))

        if args_digest in cls.Meta._query_opts_cache:
            return cls.Meta._query_opts_cache[args_digest]

        result = list()
        suboptions = dict()
        suboptions_functions = dict()

        for rela_name, rela in cls.get_relationships():  # type: str, RelationshipProperty
            include_rela = True
            include_subkeys = {
                key.replace(f'{rela_name}.', '') for key in include if key.startswith(f'{rela_name}.')
            } if isinstance(include, set) else set()
            exclude_subkeys = {
                key.replace(f'{rela_name}.', '') for key in exclude if key.startswith(f'{rela_name}.')
            } if isinstance(exclude, set) else set()

            if include:
                include_rela &= include == '*' or rela_name in include or bool(include_subkeys)
            if include_rela and exclude:
                include_rela &= not (exclude == '*' or (not exclude_subkeys and rela_name in exclude))
            if include_rela:
                options = cls.Meta.list_loader(rela.class_attribute) if rela.uselist \
                    else cls.Meta.single_loader(rela.class_attribute)

                suboptions[rela_name] = options
                new_include, new_exclude = include_subkeys or None, exclude_subkeys or None
                suboptions_function = partial(rela.entity.class_.query_options, new_include, new_exclude)
                suboptions_functions[rela_name] = suboptions_function
                result.append(options)

        cls.Meta._query_opts_cache[args_digest] = result

        for rela_name, sub_rela in suboptions.items():
            subopts = suboptions_functions[rela_name]()
            suboptions[rela_name] = sub_rela.options(*subopts)

        return result

    # noinspection PyUnresolvedReferences,PyProtectedMember
    @classmethod
    def query_for_select(cls, include_rela: Optional[Set[str]] = None,
                         exclude_rela: Optional[Set[str]] = None, **kwargs):
        query = select(cls)

        rela_map = cls.Meta.relationships.items()
        one_to_many_relationships = []

        if cls.get_engine().db_type == SqlStorageType.POSTGRES:
            for dep_key, relationship in rela_map:
                if relationship.single:
                    continue

                model_class = relationship.get_model()
                join_model_class = relationship.get_join_model()

                if model_class != join_model_class:
                    fkc = join_model_class._get_fk_columns(model_class)
                else:
                    fkc = model_class._get_id_columns()

                for pk in fkc:
                    if relationship.join_model:
                        subq = func.array_remove(func.array_agg(distinct(pk)), None)
                    else:
                        subq = func.array_remove(func.array_agg(pk), None)

                    query = query.column(subq)

                one_to_many_relationships.append((join_model_class, relationship))

        if one_to_many_relationships:
            query = query.group_by(*cls._get_id_columns())
            for model_class, relationship in one_to_many_relationships:
                query = query.outerjoin(model_class, onclause=relationship.get_on_clause(cls))

        return query.options(*cls.query_options(include_rela, exclude_rela))

    @classmethod
    def query_for_update(cls, **kwargs):
        return update(cls)

    @classmethod
    def query_for_delete(cls, **kwargs):
        return delete(cls)

    #
    # Retrieve operations
    #

    @classmethod
    async def get(cls, id_value: Any, **kwargs) -> Optional['SqlaModelBase']:
        no_cache = kwargs.pop('no_cache', False)
        async with cls._async_session_context() as ctx:
            if not no_cache:
                if instance := cls._get_from_local_cache(ctx, [id_value])[0]:
                    return instance
            query = cls.query_get(id_value, **kwargs).execution_options(timeout=ctx.sql.timeout_secs)
            if data := (await ctx.sql.connection.execute(query)).first():
                instance = cls._construct(data, **kwargs)
                await cls._register_instances(ctx, [instance])
                return await cls._build_rela(ctx, instance, **kwargs)
        return None

    @classmethod
    async def get_any(cls, indices: List[Any], **kwargs) -> List[Optional['SqlaModelBase']]:
        no_cache = kwargs.pop('no_cache', False)
        async with cls._async_session_context() as ctx:
            if not no_cache:
                objects = cls._get_from_local_cache(ctx, indices)
            else:
                objects = [None] * len(indices)
            non_cached_indices = [i for o, i in zip(objects, indices) if o is None]
            if non_cached_indices:
                query = cls.query_any(non_cached_indices, **kwargs).execution_options(timeout=ctx.sql.timeout_secs)
                data = (await ctx.sql.connection.execute(query)).all()
                if non_cached := cls._construct_list(data, **kwargs):
                    await cls._register_instances(ctx, non_cached)
                    objects += await cls._build_rela_list(ctx, non_cached, **kwargs)

        objects = {obj.get_id_value(): obj for obj in objects if obj is not None}
        return [objects[id_value] if id_value in objects else None for id_value in indices]

    @classmethod
    async def list(cls,
                   filters: Optional[List[Any]] = None,
                   sorting: Optional[List[Any]] = None,
                   skip: int = 0, limit: int = 0, params: Dict[str, Any] = None, **kwargs) -> List['SqlaModelBase']:
        query = cls.query_list(filters or [], sorting or [], skip, limit, params, **kwargs)
        async with cls._async_session_context() as ctx:
            query = query.execution_options(timeout=ctx.sql.timeout_secs)
            data = (await ctx.sql.connection.execute(query)).all()
            if objects := cls._construct_list(data, **kwargs):
                await cls._register_instances(ctx, objects)
                return await cls._build_rela_list(ctx, objects, **kwargs)
            return []

    @classmethod
    async def count(cls, filters: Optional[List[Any]] = None, params: Dict[str, Any] = None, **kwargs) -> int:
        async with cls._async_session_context() as ctx:
            query = cls.query_count(filters, params, **kwargs).execution_options(timeout=ctx.sql.timeout_secs)
            return await ctx.sql.connection.scalar(query)

    @classmethod
    async def exists(cls, id_value: Any, no_cache: bool = False, **kwargs) -> bool:
        async with cls._async_session_context() as ctx:
            if not no_cache and cls._has_in_local_cache(ctx, [id_value])[0]:
                return True
            query = cls.query_exists(id_value, **kwargs).execution_options(timeout=ctx.sql.timeout_secs)
            return bool(await ctx.sql.connection.scalar(query))

    @classmethod
    async def exists_by(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs) -> bool:
        if not filters or not isinstance(filters, list):
            return False

        async with cls._async_session_context() as ctx:
            query = cls.query_exists_by(filters, params, **kwargs).execution_options(timeout=ctx.sql.timeout_secs)
            return bool(await ctx.sql.connection.scalar(query))

    #
    # Modification operations
    #

    @classmethod
    async def create(cls, data: Union[CreateSchemaType, Dict[str, Any]], **kwargs) -> Optional['SqlaModel']:
        if isinstance(data, dict):
            create_data = data
        else:
            create_data = data.dict(exclude_unset=True)

        if create_data:
            create_data = cls.prepare_for_create(create_data)

        async with cls._async_session_context() as ctx:
            instance = cls._construct(create_data)
            ctx.sql.connection.add(instance)
            await ctx.sql.connection.flush(objects=[instance])
            await cls._register_instances(ctx, [instance])
            return await cls._build_rela(ctx, instance, **kwargs)

    async def update(self, data: Union[UpdateSchemaType, Dict[str, Any]]) -> bool:
        if isinstance(data, dict):
            update_data = data
        else:
            update_data = data.dict(exclude_unset=True)

        if update_data:
            update_data = self.prepare_for_update(update_data)

        query = self.query_get(self.get_id_value(), query=update(type(self), values=update_data))
        async with self.__class__._async_session_context() as ctx:
            query = query.execution_options(timeout=ctx.sql.timeout_secs)
            cursor = await ctx.sql.connection.execute(query)
            return cursor.rowcount == 1

    @classmethod
    async def update_by_id(cls, id_value: Any, data: Union[UpdateSchemaType, Dict[str, Any]]) -> bool:
        if isinstance(data, dict):
            update_data = data
        else:
            update_data = data.dict(exclude_unset=True)

        if update_data:
            update_data = cls.prepare_for_update(update_data)

        query = cls.query_get(id_value, query=update(cls, values=update_data))
        async with cls._async_session_context() as ctx:
            query = query.execution_options(timeout=ctx.sql.timeout_secs)
            cursor = await ctx.sql.connection.execute(query)
            return cursor.rowcount == 1

    @classmethod
    async def update_by(cls, filters: List[Any],
                        data: Union[UpdateSchemaType, Dict[str, Any]],
                        /, params: Dict[str, Any] = None, **kwargs) -> int:
        if isinstance(data, dict):
            update_data = data
        else:
            update_data = data.dict(exclude_unset=True)

        if update_data:
            update_data = cls.prepare_for_update(update_data)

        query = cls.query_update(filters, params, **kwargs).values(update_data)
        async with cls._async_session_context() as ctx:
            query = query.execution_options(timeout=ctx.sql.timeout_secs)
            cursor = await ctx.sql.connection.execute(query)
            return cursor.rowcount

    async def delete(self) -> bool:
        if instance_state(self).was_deleted:
            return False

        async with self.__class__._async_session_context() as ctx:
            await self._unregister_indices(ctx, [self.get_id_value()])
            await ctx.sql.connection.delete(self)
            await ctx.sql.connection.flush(objects=[self])
            return True

    @classmethod
    async def delete_all(cls, indices: List[Any] = None) -> bool:
        if indices is None:
            query = cls.query_for_delete()
        else:
            query = cls.query_any(indices, query=cls.query_for_delete())

        async with cls._async_session_context() as ctx:
            query = query.execution_options(timeout=ctx.sql.timeout_secs)
            cursor = await ctx.sql.connection.execute(query)
            await cls._unregister_indices(ctx, indices)
            return len(indices) == cursor.rowcount if indices is not None else cursor.rowcount > 0

    @classmethod
    async def delete_by_id(cls, id_value: Any) -> bool:
        query = cls.query_get(id_value, query=cls.query_for_delete())
        async with cls._async_session_context() as ctx:
            query = query.execution_options(timeout=ctx.sql.timeout_secs)
            cursor = await ctx.sql.connection.execute(query)
            await cls._unregister_indices(ctx, [id_value])
            return 1 == cursor.rowcount

    @classmethod
    async def delete_by(cls, filters: List[Any], /, params: Dict[str, Any] = None, **kwargs) -> int:
        query = cls.query_delete(filters, params, **kwargs)
        async with cls._async_session_context() as ctx:
            query = query.execution_options(timeout=ctx.sql.timeout_secs)
            cursor = await ctx.sql.connection.execute(query)
            return cursor.rowcount

    async def save(self, updated_fields: List[str] = None, **kwargs) -> bool:
        iss = instance_state(self)

        if iss.was_deleted:
            return False

        async with self.__class__._async_session_context() as ctx:
            if iss.transient:
                ctx.sql.connection.add(self)
            await ctx.sql.connection.flush(objects=[self])
            return True

    async def load(self, field_names: List[str] = None) -> bool:
        if instance_state(self).was_deleted:
            return False

        field_names = set(field_names) if field_names else None
        async with self.__class__._async_session_context() as ctx:
            await ctx.sql.connection.refresh(self, attribute_names=field_names)
            return True

    @classmethod
    async def save_all(cls, objects: List['SqlaModelBase'], **kwargs) -> bool:
        async with cls._async_session_context() as ctx:
            def save_all_impl(session):
                session.bulk_save_objects(objects)

            for obj in objects:
                ctx.sql.connection.add(obj)

            await ctx.sql.connection.run_sync(save_all_impl)
            await ctx.sql.connection.flush(objects=objects)
            await cls._register_instances(ctx, objects)
            return True

    #
    # Query operations
    #

    @classmethod
    def query_get(cls, id_value: Any, query=None, **kwargs):
        cls._check_abstract()
        if not isinstance(id_value, (list, tuple, dict)):
            ident_ = [id_value]
        else:
            ident_ = id_value
        columns = cls._get_id_columns()
        if len(ident_) != len(columns):
            raise ValueError(
                f'Incorrect number of values as primary key: expected {len(columns)}, got {len(ident_)}.')

        clause = query if query is not None else cls.query_for_select(**kwargs)
        for i, c in enumerate(columns):
            try:
                val = ident_[i]
            except KeyError:
                val = ident_[cls._get_field_by_column(c)]
            clause = clause.where(c == val)
        return clause

    @classmethod
    def query_any(cls, indices: List[Any], query=None, **kwargs):
        cls._check_abstract()
        columns = cls._get_id_columns()
        clause = query if query is not None else cls.query_for_select(**kwargs)
        vals_clause = []

        for ident in indices:
            if not isinstance(ident, (list, tuple, dict)):
                ident_ = [ident]
            else:
                ident_ = ident

            if len(ident_) != len(columns):
                raise ValueError(
                    f'Incorrect number of values as primary key: expected {len(columns)}, got {len(ident_)}.')

            vals = []
            for i, c in enumerate(columns):
                try:
                    val = ident_[i]
                except KeyError:
                    val = ident_[cls._get_field_by_column(c)]
                vals.append(val)

            if len(vals) == 1:
                vals_clause.append(vals[0])
            elif len(vals) > 1:
                vals_clause.append((*vals,))

        if len(columns) == 1:
            clause = clause.where(columns[columns.keys()[0]].in_(vals_clause))
        elif len(columns) > 1:
            clause = clause.where(tuple_(*columns).in_(vals_clause))
        return clause

    @classmethod
    def query_list(cls, filters: List[Any], sorting: List[Any],
                   skip: int = 0, limit: int = 0, params: Dict[str, Any] = None, **kwargs):
        query = cls.query_for_select(**kwargs).offset(skip if skip >= 0 else 0)
        if limit > 0:
            query = query.limit(limit)
        for f in filters:
            query = query.where(f)
        if filters and params:
            query = query.params(**params)
        for s in sorting:
            query = query.order_by(s)
        return query

    # noinspection PyUnresolvedReferences
    @classmethod
    def query_count(cls, filters: Optional[List[Any]] = None, params: Dict[str, Any] = None, **kwargs):
        if filters:
            query = cls.query_for_select()
            for f in filters:
                query = query.where(f)
            if params:
                query = query.params(**params)
        else:
            query = cls.__table__

        return select([func.count(literal_column('1'))]).select_from(query.alias())

    @classmethod
    def query_exists(cls, id_value: Any, **kwargs):
        return cls.query_get(id_value, query=exists(), **kwargs).select()

    @classmethod
    def query_exists_by(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs):
        query = exists(cls)

        for f in filters:
            query = query.where(f)
        if filters and params:
            query = query.params(**params)

        return query.select()

    @classmethod
    def query_update(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs):
        query = cls.query_for_update(**kwargs)

        for f in filters:
            query = query.where(f)
        if filters and params:
            query = query.params(**params)

        return query

    @classmethod
    def query_delete(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs):
        query = cls.query_for_delete(**kwargs)

        for f in filters:
            query = query.where(f)
        if filters and params:
            query = query.params(**params)

        return query

    #
    # Private
    #

    # noinspection PyUnresolvedReferences
    @classmethod
    def _check_abstract(cls):
        if cls.__table__ is None:
            raise TypeError(f'Model {cls.__name__} is abstract, no table is defined!')

    def __eq__(self, other: 'SqlaModelBase'):
        self._check_abstract()
        return type(self) == type(other) and self.get_id_value() == other.get_id_value()

    def __hash__(self):
        self._check_abstract()
        return self.get_id_value().__hash__()

    @classmethod
    def _get_id_columns(cls) -> ColumnCollection:
        return cls.__table__.primary_key.columns

    # noinspection PyProtectedMember
    @classmethod
    def _get_field_by_column(cls, c: Column) -> str:
        return cls.Meta._column_names_map.get(c.name)

    def _get_column_values(
            self, columns: ColumnCollection, force_tuple: bool = False) -> Union[Any, Tuple[Any]]:
        rv = []
        for c in columns:
            rv.append(getattr(self, self._get_field_by_column(c)))
        return rv[0] if len(rv) == 1 and not force_tuple else tuple(rv)

    @classmethod
    def _get_fk_columns(cls, model) -> ColumnCollection:
        for fk in cls.__table__.foreign_keys:
            if model.__table__ == fk.constraint.referred_table:
                return fk.constraint.columns
        return ColumnCollection()

    @classmethod
    async def _get_from_orm_cache(cls, ctx: ContextData, indices: List[Any]) -> List[Optional['SqlaModelBase']]:
        sqla_cache_indices = [identity_key(cls, id_value) for id_value in indices]

        def _get(session):
            return [session.identity_map.get(ident) for ident in sqla_cache_indices]

        return await ctx.sql.connection.run_sync(_get) if sqla_cache_indices else []

    @classmethod
    async def _attach_to_session(cls, ctx: ContextData, objects: List['SqlaModelBase']):
        states = list()

        for instance in objects:
            if instance is None:
                continue
            state = attributes.instance_state(instance)
            if state.session_id or state.key:
                continue
            state.session_id = ctx.sql.connection.sync_session.hash_key
            states.append(state)

        # noinspection PyProtectedMember
        def _set(session):
            session._register_persistent(states)

        await ctx.sql.connection.run_sync(_set)

    def _assign(self, data: Dict[str, Any]) -> 'SqlaModelBase':
        for k, v in data.items():
            try:
                attributes.set_committed_value(self, k, v)
            except KeyError:
                setattr(self, k, v)
        return self

    # noinspection PyProtectedMember,PyUnresolvedReferences
    @classmethod
    def _construct(cls, data: Dict[str, Any],
                   instance: Optional['SqlaModelBase'] = None,
                   only_fields: Optional[Set[str]] = None, **kwargs) -> Optional['SqlaModelBase']:
        column_names = [c.name for c in cls.get_columns()]
        rela_map = cls.Meta.relationships.items()
        one_to_many_models: Dict[str, int] = dict()

        if cls.get_engine().db_type == SqlStorageType.POSTGRES:
            for dep_key, relationship in rela_map:
                if relationship.single or not isinstance(relationship.key, str):
                    continue

                model_class = relationship.get_model()
                join_model_class = relationship.get_join_model()

                if model_class != join_model_class:
                    fkc = join_model_class._get_fk_columns(model_class)
                else:
                    fkc = model_class._get_id_columns()

                one_to_many_models[relationship.key] = len(fkc)

        elif isinstance(data, (tuple, Row)):
            data = data[0]

        if isinstance(data, dict):
            if instance := super()._construct(data, instance, only_fields, **kwargs):
                for k, count in one_to_many_models.items():
                    setattr(instance, k, data.get(k, []))

        elif isinstance(data, (tuple, Row)):
            if isinstance(data[0], cls):
                instance = data[0]
                data_counter = 1
            else:
                class_data = {k: v for k, v in zip(column_names, data[0:len(column_names)])}
                data_counter = len(column_names)
                instance = super()._construct(class_data, instance, only_fields, **kwargs)

            for k, count in one_to_many_models.items():
                if count == 1:
                    setattr(instance, k, data[data_counter])
                elif count > 1:
                    last_values = [None] * count
                    tuples = [data[data_counter + i] for i in range(count)]
                    for tuple_ in itertools.zip_longest(*tuples, fillvalue=None):
                        for i, v in enumerate(tuple_):
                            if v is None and last_values[i] is not None:
                                tuples[i].append(last_values[i])
                            else:
                                last_values[i] = v
                    setattr(instance, k, list(zip(*tuples)))
                data_counter += count
        else:
            instance = data

        return instance

    @classmethod
    async def _register_instances(cls, ctx: ContextData, objects: List['SqlaModelBase']):
        await super()._register_instances(ctx, objects)
        await cls._attach_to_session(ctx, objects)



import contextlib
from collections import defaultdict
from dataclasses import dataclass, replace
from typing import Any, Dict, List, Optional, Set, Tuple

from dropland.blocks.sql.base import SqlStorageType
from dropland.data.context import get_context
from .engine import EngineKey, SqlStorageEngine, SqlStorageBackend


@dataclass
class Session:
    engine: SqlStorageEngine
    connection: Any
    timeout_secs: int
    in_transaction: bool = False


class ConnectionContext:
    def __init__(self):
        self.sessions: Dict[str, Session] = dict()
        self.sessions_by_key: Dict[EngineKey, Set[str]] = defaultdict(set)


class SessionManager:
    def __init__(self, engine_factory: SqlStorageBackend):
        self._engine_factory = engine_factory

    def get_session(self, name: str) -> Optional[Session]:
        ctx = get_context()
        if '_sql_session' not in ctx:
            ctx._sql_session = ConnectionContext()
        return ctx._sql_session.sessions.get(name)

    def get_sessions_for_type(self, db_type: SqlStorageType, is_async: bool) -> List[Session]:
        ctx = get_context()
        if '_sql_session' not in ctx:
            ctx._sql_session = ConnectionContext()
        conn_names = ctx._sql_session.sessions_by_key[EngineKey(db_type=db_type, is_async=is_async)]
        return [ctx._sql_session.sessions.get(name) for name in conn_names if name in ctx._sql_session.sessions]

    def get_or_create_session(self, name: str) -> Tuple[bool, Optional[Session]]:
        if conn := self.get_session(name):
            return False, conn

        if engine := self._engine_factory.get_engine(name):
            return True, Session(
                engine=engine, connection=engine.new_connection(),
                timeout_secs=int(engine.timeout.total_seconds()))

        return False, None

    @contextlib.contextmanager
    def sync_session_context(self, name: str, begin_tx: bool = True, autocommit: bool = True):
        created, session = self.get_or_create_session(name)
        assert not session.engine.is_async, \
            f'Sql engine with name "{name}" has only async driver, ' \
            f'use async_session_context() function instead'

        if not created:
            if begin_tx and not session.in_transaction:
                with session.engine.sync_transaction_context(session.connection, autocommit):
                    yield self._add_session(name, replace(session, in_transaction=True))
                    self._add_session(name, replace(session, in_transaction=False))
            else:
                yield session
            return

        with session.connection as conn:
            assert isinstance(conn, session.engine.connection_class), \
                f'Engine with name "{name}" has only async driver, ' \
                f'use async_session_context() function instead'

            session = replace(session, connection=conn, in_transaction=begin_tx)
            if begin_tx:
                with session.engine.sync_transaction_context(session.connection, autocommit):
                    yield self._add_session(name, session)
            else:
                yield self._add_session(name, session)
            self._remove_session(name)

    @contextlib.asynccontextmanager
    async def session_context(self, name: str, begin_tx: bool = True, autocommit: bool = True):
        created, session = self.get_or_create_session(name)
        assert session.engine.is_async, \
            f'Sql engine with name "{name}" has only sync driver, ' \
            f'use session_context() function instead'

        if not created:
            if begin_tx and not session.in_transaction:
                async with session.engine.transaction_context(session.connection, autocommit):
                    yield self._add_session(name, replace(session, in_transaction=True))
                    self._add_session(name, replace(session, in_transaction=False))
            else:
                yield session
            return

        async with session.connection as conn:
            assert isinstance(conn, session.engine.connection_class), \
                f'Sql engine with name "{name}" has only sync driver, ' \
                f'use session_context() function instead'

            session = replace(session, connection=conn, in_transaction=begin_tx)
            if begin_tx:
                async with session.engine.transaction_context(session.connection, autocommit):
                    yield self._add_session(name, session)
            else:
                yield self._add_session(name, session)
            self._remove_session(name)

    @contextlib.contextmanager
    def init_sync_engines(self, names: List[str] = None):
        engines = self._engine_factory.get_engines(names or [])
        with contextlib.ExitStack() as stack:
            for name, engine in engines.items():
                if engine.is_async:
                    continue
                engine.sync_start()
                stack.callback(engine.sync_stop)
            yield

    @contextlib.asynccontextmanager
    async def init_engines(self, names: List[str] = None):
        engines = self._engine_factory.get_engines(names or [])
        async with contextlib.AsyncExitStack() as stack:
            for name, engine in engines.items():
                if engine.is_async:
                    await engine.start()
                    stack.push_async_callback(engine.stop)
                else:
                    engine.sync_start()
                    stack.callback(engine.sync_stop)
            yield

    def _add_session(self, name: str, data: Session) -> Session:
        ctx = get_context()
        if '_sql_session' not in ctx:
            ctx._sql_session = ConnectionContext()
        ctx._sql_session.sessions[name] = data
        ctx._sql_session.sessions_by_key[
            EngineKey(db_type=data.engine.db_type, is_async=data.engine.is_async)
        ].add(name)
        return data

    def _remove_session(self, name: str):
        ctx = get_context()
        if '_sql_session' in ctx:
            data = ctx._sql_session.sessions.pop(name, None)
            ctx._sql_session.sessions_by_key.pop(
                EngineKey(db_type=data.engine.db_type, is_async=data.engine.is_async), None)



import pytest

from dropland.blocks.redis import USE_REDIS

pytestmark = pytest.mark.skipif(not USE_REDIS, reason='For Redis only')

if USE_REDIS:
    from dropland.blocks.redis.model import RedisCacheType, RedisModel
    from .redis_models_data import redis_engine


    class TestModel(RedisModel, redis_engine=redis_engine, cache_type=RedisCacheType.HASH):
        a: int = 1
        b: str = '123'

        def __init__(self):
            self.a = 1
            self.b = '123'

        def get_id_value(self):
            return self.a


@pytest.mark.asyncio
async def test_model_fields(redis_session):
    assert TestModel.get_model_cache_key() == 'dropland.models.TestModel'
    assert TestModel.get_cache_key(1) == 'dropland.models.TestModel:1'
    assert TestModel.get_cache_id(123) == '123'
    assert TestModel.get_engine() is redis_engine


@pytest.mark.asyncio
async def test_model_serde(redis_session):
    m = TestModel()
    assert m.a == 1
    assert m.b == '123'

    assert TestModel.get_fields() == {'a', 'b'}
    assert m.get_values() == {'a': 1, 'b': '123'}

    ser = m.serialize()
    assert isinstance(ser, bytes)
    res = TestModel.deserialize(ser)
    assert isinstance(res, TestModel)

    assert res.a == 1
    assert res.b == '123'


@pytest.mark.asyncio
async def test_exists_save_and_get(redis_session):
    m = TestModel()

    assert not await TestModel.exists(1)
    assert not await TestModel.get(1)

    assert await m.save()
    assert await TestModel.exists(1)

    m2 = await TestModel.get(1)
    assert m2.a == 1
    assert m2.b == '123'

    await m.delete()
    assert not await TestModel.exists(1)
    assert not await TestModel.get(1)


@pytest.mark.asyncio
async def test_load(redis_session):
    m = TestModel()

    assert not await m.load()
    assert await m.save()
    assert await m.load()

    assert m.a == 1
    assert m.b == '123'

    m.b = None
    assert m.a == 1
    assert m.b is None

    assert await m.load()
    assert m.a == 1
    assert m.b == '123'


@pytest.mark.asyncio
async def test_delete(redis_session):
    m = TestModel()
    await m.save()

    m2 = await TestModel.get(1)
    assert m2.a == 1
    assert m2.b == '123'

    assert await m.load()
    assert m.a == 1
    assert m.b == '123'

    assert await TestModel.exists(1)
    assert await m.delete()
    assert not await m.delete()

    assert not await TestModel.exists(1)
    assert not await TestModel.get(1)
    assert not await m.load()


@pytest.mark.asyncio
async def test_save_all_get_any_and_scan(redis_session):
    m, m2 = TestModel(), TestModel()
    m2.a = 2

    assert await TestModel.save_all([m, m2])

    objects = await TestModel.get_any([])
    assert len(objects) == 0

    objects = await TestModel.get_any([1, 2])
    assert len(objects) == 2
    assert objects[0].a == 1
    assert objects[1].a == 2

    objects = await TestModel.get_any([-1, 1, 999999, 2, 0])
    assert len(objects) == 5
    assert objects[0] is None
    assert objects[1].a == 1
    assert objects[2] is None
    assert objects[3].a == 2
    assert objects[4] is None

    objects = {k: v async for k, v in TestModel.scan('*')}
    assert len(objects) == 2
    assert objects['1'].a == 1
    assert objects['2'].a == 2


@pytest.mark.asyncio
async def test_delete_all(redis_session):
    m, m2 = TestModel(), TestModel()
    m2.a = 2

    assert await TestModel.save_all([m, m2])
    objects = await TestModel.get_any([1, 2])
    assert len(objects) == 2
    assert objects[0].a == 1
    assert objects[1].a == 2

    assert await TestModel.delete_all([1, 2])
    assert not await TestModel.delete_all([1, 2])

    assert await TestModel.save_all([m, m2])
    assert await TestModel.delete_all()
    assert not await TestModel.delete_all()

    objects = await TestModel.get_any([1, 2])
    assert objects == [None, None]



import asyncio
import contextlib
import enum
from datetime import timedelta
from typing import Any, Optional

from dependency_injector.wiring import Provide, inject

from dropland.data.context import get_context
from dropland.data.models.cache import MethodCache
from dropland.data.models.nosql import NoSqlModel, NoSqlProxyModel
from dropland.data.serializers import Serializer, Deserializer
from dropland.data.serializers.pickle import PickleSerializer, PickleDeserializer
from .engine import RedisStorageEngine


class RedisCacheType(int, enum.Enum):
    SIMPLE = 0
    HASH = 1


class RedisModel(NoSqlModel):
    class Meta(NoSqlModel.Meta):
        _redis_engine = None

    def __init_subclass__(
            cls, redis_engine: RedisStorageEngine = None,
            cache_type: RedisCacheType = RedisCacheType.SIMPLE,
            serializer: Optional[Serializer] = None,
            deserializer: Optional[Deserializer] = None,
            ttl_enabled: bool = True,
            **kwargs):
        from .cache import SimpleRedisModelCache, HashRedisModelCache

        if not redis_engine:
            return super().__init_subclass__(**kwargs)

        serializer = serializer or PickleSerializer()
        deserializer = deserializer or PickleDeserializer()

        protocol_class = SimpleRedisModelCache \
            if cache_type == RedisCacheType.SIMPLE \
            else HashRedisModelCache

        cache_protocol = protocol_class(
            redis_engine, cls.__name__,
            serializer, deserializer,
            ttl_enabled=ttl_enabled
        )

        cls.Meta._redis_engine = redis_engine

        super().__init_subclass__(
            cache_protocol=cache_protocol,
            serializer=serializer,
            deserializer=deserializer,
            **kwargs
        )

    def get_id_value(self) -> Any:
        raise NotImplementedError

    # noinspection PyProtectedMember
    @classmethod
    def get_engine(cls) -> 'RedisStorageEngine':
        return cls.Meta._redis_engine

    @classmethod
    @contextlib.asynccontextmanager
    @inject
    async def _async_session_context(cls, redis_block=Provide['block']):
        ctx = get_context()

        if '_redis_cond' not in ctx:
            ctx._redis_cnt = 0
            ctx._redis_cond = asyncio.Condition()

        async with ctx._redis_cond:
            if 'redis' in ctx:
                ctx._redis_cnt += 1
                ctx._redis_cond.release()
                try:
                    yield ctx
                finally:
                    await ctx._redis_cond.acquire()
                    ctx._redis_cnt -= 1
                    if 0 == ctx._redis_cnt:
                        ctx._redis_cond.notify()
            else:
                engine_name = cls.get_engine().name
                async with redis_block.container.manager().session_context(engine_name):
                    ctx.redis = redis_block.get_session(engine_name)
                    assert ctx.redis
                    ctx._redis_cnt = 1
                    ctx._redis_cond.release()
                    try:
                        yield ctx
                    finally:
                        await ctx._redis_cond.acquire()
                        ctx._redis_cnt -= 1
                        if 0 == ctx._redis_cnt:
                            del ctx.redis
                        else:
                            await ctx._redis_cond.wait_for(lambda: 0 == ctx._redis_cnt)
                            del ctx.redis


class RedisProxyModel(NoSqlProxyModel, RedisModel):
    class Meta(NoSqlProxyModel.Meta, RedisModel.Meta):
        pass

    def __init_subclass__(cls, redis_engine: RedisStorageEngine = None, **kwargs):
        super().__init_subclass__(redis_engine=redis_engine, **kwargs)
        cls.Meta._redis_engine = redis_engine

    def get_id_value(self) -> Any:
        raise NotImplementedError


class RedisMethodCache(MethodCache):
    def __init__(
            self, redis_engine: RedisStorageEngine, model_name: str,
            cache_type: RedisCacheType = RedisCacheType.SIMPLE,
            serializer: Optional[Serializer] = None,
            deserializer: Optional[Deserializer] = None,
            ttl: Optional[timedelta] = None):
        from .cache import SimpleRedisModelCache, HashRedisModelCache

        serializer = serializer or PickleSerializer()
        deserializer = deserializer or PickleDeserializer()

        protocol_class = SimpleRedisModelCache \
            if cache_type == RedisCacheType.SIMPLE \
            else HashRedisModelCache

        cache_protocol = protocol_class(
            redis_engine, model_name,
            serializer, deserializer,
            ttl_enabled=True
        )

        super().__init__(cache_protocol, ttl)
        self._redis_engine = redis_engine
        self._cache_type = cache_type
        self._engine_name = redis_engine.name

    @contextlib.asynccontextmanager
    @inject
    async def _async_session_context(self, redis_block=Provide['block']):
        ctx = get_context()

        if '_redis_cond' not in ctx:
            ctx._redis_cnt = 0
            ctx._redis_cond = asyncio.Condition()

        async with ctx._redis_cond:
            if 'redis' in ctx:
                ctx._redis_cnt += 1
                ctx._redis_cond.release()
                try:
                    yield ctx
                finally:
                    await ctx._redis_cond.acquire()
                    ctx._redis_cnt -= 1
                    if 0 == ctx._redis_cnt:
                        ctx._redis_cond.notify()
            else:
                async with redis_block.container.manager().session_context(self._engine_name):
                    ctx.redis = redis_block.get_session(self._engine_name)
                    assert ctx.redis
                    ctx._redis_cnt = 1
                    ctx._redis_cond.release()
                    try:
                        yield ctx
                    finally:
                        await ctx._redis_cond.acquire()
                        ctx._redis_cnt -= 1
                        if 0 == ctx._redis_cnt:
                            del ctx.redis
                        else:
                            await ctx._redis_cond.wait_for(lambda: 0 == ctx._redis_cnt)
                            del ctx.redis



import asyncio
import contextlib
import operator
from datetime import timedelta
from functools import reduce
from typing import Dict, Any, Optional, Tuple, List

from dropland.util import calculate_digest
from ..cache import ModelCacheProtocol, ModelCacheData


class MethodCache:
    def __init__(self, cache_protocol: ModelCacheProtocol, ttl: Optional[timedelta] = None):
        self._cache_protocol = cache_protocol
        self._ttl = ttl

    @property
    def cache_protocol(self) -> ModelCacheProtocol:
        return self._cache_protocol

    def get_model_cache_key(self) -> str:
        return self._cache_protocol.get_model_cache_key()

    def get_cache_id(self, id_value: Any) -> str:
        return self._cache_protocol.get_cache_id(id_value)

    def get_cache_key(self, id_value: Any) -> str:
        return self._cache_protocol.get_cache_key(id_value)

    @contextlib.asynccontextmanager
    async def _async_session_context(self, session=None):
        ...

    @staticmethod
    def cache_key_for(method: str, args: Optional[Dict[str, Any]] = None) -> str:
        id_data = [method]

        if args:
            id_data.append(calculate_digest(args))
        else:
            id_data.append('_')

        return ':'.join(id_data)

    async def put(self, method: str, args: Optional[Dict[str, Any]] = None,
                  data: Optional[Any] = None, cache_key: Optional[str] = None,
                  ttl: Optional[timedelta] = None, **kwargs) -> bool:
        cache_key = cache_key or self.cache_key_for(method, args)
        async with self._async_session_context() as ctx:
            data = ModelCacheData(cache_id=cache_key, data=data)
            return await self._cache_protocol.cache_one(ctx, data, ttl=ttl or self._ttl, **kwargs)

    async def get(self, method: str, args: Optional[Dict[str, Any]] = None,
                  cache_key: Optional[str] = None, **kwargs) -> Tuple[bool, Optional[Any]]:
        cache_key = cache_key or self.cache_key_for(method, args)
        async with self._async_session_context() as ctx:
            exists, data = await self._cache_protocol.load_one(
                ctx, self._cache_protocol.get_cache_key(cache_key), **kwargs)
            return exists, data

    async def drop(self, method: str, args: Optional[Dict[str, Any]] = None,
                   cache_key: Optional[str] = None) -> bool:
        cache_key = cache_key or self.cache_key_for(method, args)
        async with self._async_session_context() as ctx:
            return await self._cache_protocol.drop_one(ctx, self._cache_protocol.get_cache_key(cache_key))

    async def drop_all(self, methods: List[str] = None) -> int:
        methods = methods if methods else list()
        async with self._async_session_context() as ctx:
            if not methods:
                return await self._cache_protocol.drop_all(ctx)
            results = await asyncio.gather(
                *(self._cache_protocol.drop_all(ctx, method) for method in methods)
            )
            return reduce(operator.add, results, 0)



import pytest

from tests.sql_models_data import USE_SQL

pytestmark = pytest.mark.skipif(not USE_SQL, reason='For SqlAlchemy only')

if USE_SQL:
    from dropland.blocks.sql.base import SqlStorageType
    from dropland.blocks.sql.containers import SqlStorage, SingleSqlStorage, MultipleSqlStorage
    from dropland.blocks.sql.engine import EngineConfig, SqlStorageEngine
    from tests import MYSQL_URI, POSTGRES_URI, SQLITE_URI


@pytest.mark.asyncio
async def test_create_engine():
    sqlite_config = EngineConfig(
        url=SQLITE_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )
    pg_config = EngineConfig(
        url=POSTGRES_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )
    mysql_config = EngineConfig(
        url=MYSQL_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    default_sql_storage = SqlStorage()
    engine_factory = default_sql_storage.engine_factory()

    assert engine_factory.create_engine('sqlite', sqlite_config, SqlStorageType.SQLITE, use_async=False)
    assert engine_factory.create_engine('asqlite', sqlite_config, SqlStorageType.SQLITE, use_async=True)
    assert engine_factory.create_engine('pg', pg_config, SqlStorageType.POSTGRES, use_async=False)
    assert engine_factory.create_engine('apg', pg_config, SqlStorageType.POSTGRES, use_async=True)
    assert engine_factory.create_engine('ms', mysql_config, SqlStorageType.MYSQL, use_async=False)
    assert engine_factory.create_engine('ams', mysql_config, SqlStorageType.MYSQL, use_async=True)

    for db_type in (SqlStorageType.SQLITE, SqlStorageType.POSTGRES, SqlStorageType.MYSQL):
        for e in engine_factory.get_engines_for_type(db_type, is_async=False):
            assert e.db_type == db_type
            assert not e.is_async

        for e in engine_factory.get_engines_for_type(db_type, is_async=True):
            assert e.db_type == db_type
            assert e.is_async

    assert engine_factory.get_engine_names() == ['sqlite', 'asqlite', 'pg', 'apg', 'ms', 'ams']


def test_sqlite_engine(sqlite_engine):
    sqlite_engine.start()

    with sqlite_engine.new_connection() as conn:
        res = conn.execute('select sqlite_version();')
        print(res.fetchone()[0])
        res = conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    sqlite_engine.stop()


@pytest.mark.asyncio
async def test_async_sqlite_engine(sqlite_async_engine):
    await sqlite_async_engine.async_start()

    async with sqlite_async_engine.new_connection() as conn:
        res = await conn.execute('select sqlite_version();')
        print(res.fetchone()[0])
        res = await conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    await sqlite_async_engine.async_stop()


def test_pg_engine(pg_engine):
    pg_engine.start()

    with pg_engine.new_connection() as conn:
        res = conn.execute('select version();')
        print(res.fetchone()[0])
        res = conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    pg_engine.stop()


@pytest.mark.asyncio
async def test_async_pg_engine(pg_async_engine):
    await pg_async_engine.async_start()

    async with pg_async_engine.new_connection() as conn:
        res = await conn.execute('select version();')
        print(res.fetchone()[0])
        res = await conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    await pg_async_engine.async_stop()


def test_mysql_engine(mysql_engine):
    mysql_engine.start()

    with mysql_engine.new_connection() as conn:
        res = conn.execute('select version();')
        print(res.fetchone()[0])
        res = conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    mysql_engine.stop()


@pytest.mark.asyncio
async def test_async_mysql_engine(mysql_async_engine):
    await mysql_async_engine.async_start()

    async with mysql_async_engine.new_connection() as conn:
        res = await conn.execute('select version();')
        print(res.fetchone()[0])
        res = await conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    await mysql_async_engine.async_stop()


@pytest.mark.asyncio
async def test_storage_container():
    sqlite_config = EngineConfig(
        url=SQLITE_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    cont = SqlStorage()

    eng = cont.create_engine('dropland', sqlite_config, SqlStorageType.SQLITE, use_async=False)
    assert isinstance(eng, SqlStorageEngine)
    assert eng.name == 'dropland'
    cont.unwire()

    cont = SingleSqlStorage()
    cont.config.from_dict({
        'name': '_',
        'db_type': SqlStorageType.SQLITE,
        'engine_config': sqlite_config,
        'use_async': False,
    })
    eng1 = cont.create_engine()
    eng2 = cont.create_engine()
    assert eng1.name == eng2.name == '_'
    assert eng1 is eng2
    cont.unwire()

    cont = MultipleSqlStorage()
    cont.config.from_dict({
        'one': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': sqlite_config,
            'use_async': False,
        },
        'two': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': sqlite_config,
            'use_async': True,
        },
    })

    eng1 = cont.create_engine('one')
    eng2 = cont.create_engine('two')
    assert eng1 is not eng2
    assert eng1.is_async is False
    assert eng2.is_async is True
    assert eng1.name == 'one'
    assert eng2.name == 'two'

    cont.unwire()



from datetime import datetime
from typing import Union

import rpyc
from apscheduler.schedulers.base import STATE_STOPPED
from apscheduler.util import undefined
from dependency_injector import containers
from dependency_injector.containers import Container
from dependency_injector.wiring import Provide, inject

from dropland.app.application import Application
from dropland.blocks.scheduler.containers import SchedulerBlock, SimpleSchedulerContainer
from dropland.log import tr, worker_logger
from .engine import EngineConfig
from .local import Scheduler


class SchedulerRPC(rpyc.Service):
    def __init__(self, scheduler: Scheduler):
        self._scheduler = scheduler

    def _prepare_args(self, args: tuple) -> tuple:
        result = []
        for value in args:
            if isinstance(value, datetime):
                value = value.strftime('%Y-%m-%dT%H:%M:%S.%f')
            elif isinstance(value, type(undefined)):
                value = undefined
            result.append(value)
        return tuple(result)

    def _prepare_kwargs(self, args: dict) -> dict:
        for k, v in args.items():
            if isinstance(v, datetime):
                args[k] = v.strftime('%Y-%m-%dT%H:%M:%S.%f')
            elif isinstance(v, type(undefined)):
                args[k] = undefined
        return args

    def exposed_start(self, paused=False):
        return self._scheduler.start(paused)

    def exposed_shutdown(self, wait=True):
        return self._scheduler.shutdown(wait)

    def exposed_pause(self):
        return self._scheduler.pause()

    def exposed_resume(self):
        return self._scheduler.resume()

    def exposed_add_job(self, func, *args, **kwargs):
        return self._scheduler.add_job(func, *self._prepare_args(args), **self._prepare_kwargs(kwargs))

    def exposed_scheduled_job(self, trigger, *args, **kwargs):
        return self._scheduler.scheduled_job(trigger, *args, **kwargs)

    def exposed_modify_job(self, job_id, jobstore=None, **changes):
        if _ := self._scheduler.get_job(job_id, jobstore):
            self._scheduler.modify_job(job_id, jobstore, **changes)
            return True
        return False

    def exposed_reschedule_job(self, job_id, jobstore=None, trigger=None, **trigger_args):
        return self._scheduler.reschedule_job(job_id, jobstore, trigger, **trigger_args)

    def exposed_pause_job(self, job_id, jobstore=None):
        return self._scheduler.pause_job(job_id, jobstore)

    def exposed_resume_job(self, job_id, jobstore=None):
        return self._scheduler.resume_job(job_id, jobstore)

    def exposed_remove_job(self, job_id, jobstore=None):
        return self._scheduler.remove_job(job_id, jobstore)

    def exposed_remove_all_jobs(self, jobstore=None):
        return self._scheduler.remove_all_jobs(jobstore)

    def exposed_get_job(self, job_id, jobstore=None):
        return self._scheduler.get_job(job_id, jobstore)

    def exposed_get_jobs(self, jobstore=None, pending=None):
        return self._scheduler.get_jobs(jobstore, pending)

    @property
    def exposed_running(self):
        return self._scheduler.running


class SchedulerApplication(Application):
    def __init__(self, container: Container, name: str,
                 engine_config: Union[EngineConfig, dict],
                 scheduler_block_name: str = 'scheduler'):
        super().__init__(container, name)

        scheduler_block: SchedulerBlock = self.get_resource(scheduler_block_name)
        scheduler_block.container.as_service.override(True)
        scheduler_block.container.reset_singletons()

        self._scheduler = scheduler_block.get_instance()
        self._engine_config = engine_config

    @property
    def scheduler(self):
        return self._scheduler

    def _create_server(self, engine_config: Union[EngineConfig, dict]):
        from rpyc.utils.server import ThreadPoolServer

        worker_logger.info(tr('dropland.blocks.scheduler.service.start'))

        engine_config = engine_config if isinstance(engine_config, EngineConfig) else EngineConfig(**engine_config)

        config = rpyc.core.protocol.DEFAULT_CONFIG.copy()
        config.update({'allow_public_attrs': True, 'allow_pickle': True, 'import_custom_exceptions': True})

        return ThreadPoolServer(
            SchedulerRPC(self._scheduler), nbThreads=engine_config.task_workers,
            hostname=engine_config.task_host, port=engine_config.task_port,
            protocol_config=config
        )

    # noinspection PyBroadException
    def run(self):
        try:
            server = self._create_server(self._engine_config)
            server.start()

        except (KeyboardInterrupt, SystemExit):
            pass
        except BaseException:
            worker_logger.exception(tr('dropland.blocks.scheduler.service.exception'))
        finally:
            worker_logger.info(tr('dropland.blocks.scheduler.service.stop'))
            if STATE_STOPPED != self._scheduler.state:
                self._scheduler.shutdown()


@inject
def run_service(container: SimpleSchedulerContainer = Provide['<container>']):
    """ Worker service entrypoint """
    if not isinstance(container, (SimpleSchedulerContainer, containers.DynamicContainer)):
        container = SimpleSchedulerContainer()
        application = SchedulerApplication(container, 'worker', EngineConfig())
    else:
        application = container.instance(container)

    application.run()


if __name__ == '__main__':
    run_service()



import os
from dataclasses import dataclass
from datetime import timedelta
from typing import Dict, List, Optional, Union

import pytz
from apscheduler.schedulers.base import BaseScheduler

from dropland.blocks.redis import USE_REDIS
from dropland.blocks.sql import SqlStorageEngine
from dropland.log import logger, tr


@dataclass
class EngineConfig:
    sql_url: Union[str, SqlStorageEngine] = None
    sql_tablename: Optional[str] = None
    redis_url: Optional[str] = None
    redis_job_key: Optional[str] = None

    job_coalesce: bool = False
    job_max_instances: int = 1
    job_misfire_grace_time: int = 24 * 3600

    task_host: str = '0.0.0.0'
    task_port: int = 3000
    task_processes: int = os.cpu_count()
    task_workers: int = os.cpu_count()
    task_rpc_timeout_seconds: int = 5
    task_rpc_num_connect_attempts: int = 10
    create_remote_engine: Optional[bool] = None
    timezone: str = 'UTC'


class SchedulerBackend:
    def __init__(self, as_service: bool):
        self._engines: Dict[str, BaseScheduler] = dict()
        self._as_service = as_service

    @property
    def name(self) -> str:
        return 'scheduler'

    def create_engine(self, name: str, config: EngineConfig) -> Optional[BaseScheduler]:
        if engine := self._engines.get(name):
            return engine

        from apscheduler.jobstores.memory import MemoryJobStore

        jobstores = {'default': MemoryJobStore()}

        if USE_REDIS and config.redis_url:
            from apscheduler.jobstores.redis import RedisJobStore
            from redis import ConnectionPool

            jobstores['redis'] = RedisJobStore(
                connection_pool=ConnectionPool.from_url(config.redis_url),
                jobs_key=f'{config.redis_job_key}.aps.jobs'
                    if isinstance(config.redis_job_key, str) else 'apscheduler.jobs',
                run_times_key=f'{config.redis_job_key}.aps.run_times'
                    if isinstance(config.redis_job_key, str) else 'apscheduler.run_times',
            )

        if config.sql_url:
            from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore

            jobstores['sql'] = SQLAlchemyJobStore(
                url=config.sql_url if isinstance(config.sql_url, str) else None,
                engine=config.sql_url if not isinstance(config.sql_url, str) else None,
                metadata=config.sql_url.metadata if not isinstance(config.sql_url, str) else None,
                tablename=config.sql_tablename if config.sql_tablename else 'apscheduler_jobs',
                engine_options={'pool_pre_ping': True},
            )

        job_defaults = {
            'coalesce': config.job_coalesce,
            'max_instances': config.job_max_instances,
            'misfire_grace_time': config.job_misfire_grace_time,
        }

        engine = self._create_engine(config, jobstores=jobstores, job_defaults=job_defaults)
        self._engines[name] = engine
        logger.info(tr('dropland.blocks.scheduler.created'))
        return engine

    def get_engine(self, name: str) -> Optional[BaseScheduler]:
        return self._engines.get(name)

    def get_engine_names(self) -> List[str]:
        return list(self._engines.keys())

    def _create_engine(self, config: EngineConfig, **kwargs):
        from apscheduler.executors.asyncio import AsyncIOExecutor
        from apscheduler.executors.pool import ProcessPoolExecutor
        from concurrent.futures.thread import ThreadPoolExecutor as BackgroundExecutor
        from .local import Scheduler

        logger.info(tr('dropland.blocks.scheduler.create.local'))

        executors = {
            'default': AsyncIOExecutor(),
            'process': ProcessPoolExecutor(config.task_processes),
        }

        if config.timezone:
            timezone = pytz.timezone(config.timezone) if isinstance(config.timezone, str) else config.timezone
        else:
            timezone = None

        create_remote_engine = config.create_remote_engine and not self._as_service
        executor = BackgroundExecutor(config.task_workers, thread_name_prefix='TaskWorker')
        scheduler = Scheduler(
            remote=self._create_remote(config) if create_remote_engine else None,
            executor=executor, executors=executors, timezone=timezone, **kwargs
        )
        return scheduler

    def _create_remote(self, config: EngineConfig):
        import apscheduler.jobstores.base
        from .remote import RemoteScheduler, register_exception

        logger.info(tr('dropland.blocks.scheduler.create.remote'))

        register_exception(apscheduler.jobstores.base.JobLookupError)
        register_exception(apscheduler.jobstores.base.ConflictingIdError)
        register_exception(apscheduler.jobstores.base.TransientJobError)

        return RemoteScheduler(
            host=config.task_host, port=config.task_port,
            num_connect_attempts=config.task_rpc_num_connect_attempts,
            rpc_timeout=timedelta(seconds=config.task_rpc_timeout_seconds)
        )



from dropland.blocks.redis import USE_REDIS

if USE_REDIS:
    from dropland.blocks.redis.containers import RedisStorage
    from dropland.blocks.redis.engine import EngineConfig

from tests import REDIS_URI

if USE_REDIS:
    redis_engine = RedisStorage().create_engine('dropland', EngineConfig(url=REDIS_URI))



import contextlib
from datetime import timedelta
from typing import Optional, Any, Tuple, List, Union, Dict

from .model import Model, CreateSchemaType, UpdateSchemaType
from .serializable import SerializableModel
from ..cache import ModelCacheProtocol, ModelCacheData
from ..serializers import Serializer, Deserializer


class NoSqlModel(Model, SerializableModel):
    class Meta(Model.Meta):
        cache_protocol: ModelCacheProtocol

    def __init_subclass__(
            cls, cache_protocol: Optional[ModelCacheProtocol] = None,
            serializer: Optional[Serializer] = None,
            deserializer: Optional[Deserializer] = None,
            **kwargs):
        super().__init_subclass__(serializer, deserializer, **kwargs)
        cls.Meta.cache_protocol = cache_protocol

    @classmethod
    def get_model_cache_key(cls) -> str:
        return cls.Meta.cache_protocol.get_model_cache_key()

    @classmethod
    def get_cache_id(cls, id_value: Any) -> str:
        return cls.Meta.cache_protocol.get_cache_id(id_value)

    @classmethod
    def get_cache_key(cls, id_value: Any) -> str:
        return cls.Meta.cache_protocol.get_cache_key(id_value)

    @classmethod
    @contextlib.asynccontextmanager
    async def _async_session_context(cls):
        ...

    #
    # Retrieve operations
    #

    @classmethod
    async def get(cls, id_value: Any, **kwargs) -> Optional['NoSqlModel']:
        no_cache = kwargs.pop('no_cache', False)
        async with cls._async_session_context() as ctx:
            if not no_cache:
                if instance := cls._get_from_local_cache(ctx, [id_value])[0]:
                    return instance
            exists, data = await cls.Meta.cache_protocol.load_one(
                ctx, cls.get_cache_key(id_value), **kwargs)
            if exists:
                if instance := cls._construct(data, **kwargs):
                    await cls._register_instances(ctx, [instance])
                    return await cls._build_rela(ctx, instance, **kwargs)
        return None

    @classmethod
    async def get_any(cls, indices: List[Any], **kwargs) -> List[Optional['NoSqlModel']]:
        no_cache = kwargs.pop('no_cache', False)
        async with cls._async_session_context() as ctx:
            if not no_cache:
                objects = cls._get_from_local_cache(ctx, indices)
            else:
                objects = [None] * len(indices)
            non_cached_indices = [i for o, i in zip(objects, indices) if o is None]
            if non_cached_indices:
                data = await cls.Meta.cache_protocol.load_many(ctx, non_cached_indices, **kwargs)
                if non_cached := cls._construct_list(data, **kwargs):
                    await cls._register_instances(ctx, non_cached)
                    objects += await cls._build_rela_list(ctx, non_cached, **kwargs)

        objects = {obj.get_id_value(): obj for obj in objects if obj is not None}
        return [objects[id_value] if id_value in objects else None for id_value in indices]

    @classmethod
    async def exists(cls, id_value: Any, no_cache: bool = False) -> bool:
        async with cls._async_session_context() as ctx:
            if not no_cache and cls._has_in_local_cache(ctx, [id_value])[0]:
                return True
            return await cls.Meta.cache_protocol.exists(ctx, cls.get_cache_key(id_value))

    @classmethod
    async def scan(cls, match: str, count: int = None, **kwargs) -> Tuple[str, Optional['NoSqlModel']]:
        async with cls._async_session_context() as ctx:
            async for k, data in cls.Meta.cache_protocol.scan(ctx, cls.get_model_cache_key(), match, count):
                if data is None:
                    yield k, None
                elif instance := cls._construct(data, **kwargs):
                    await cls._register_instances(ctx, [instance])
                    yield k, await cls._build_rela(ctx, instance, **kwargs)
                else:
                    yield k, None

    #
    # Modification operations
    #

    @classmethod
    async def create(cls, data: Union[CreateSchemaType, Dict[str, Any]], **kwargs) -> Optional['NoSqlModel']:
        if isinstance(data, dict):
            create_data = data
        else:
            create_data = data.dict(exclude_unset=True)

        instance = cls._construct(create_data)
        await instance.save()
        return instance

    async def update(self, data: Union[UpdateSchemaType, Dict[str, Any]]) -> bool:
        if isinstance(data, dict):
            update_data = data
        else:
            update_data = data.dict(exclude_unset=True)

        instance = self._construct(update_data, instance=self)
        return await instance.save()

    @classmethod
    async def update_by_id(cls, id_value: Any, data: Union[UpdateSchemaType, Dict[str, Any]]) -> bool:
        if isinstance(data, dict):
            update_data = data
        else:
            update_data = data.dict(exclude_unset=True)

        async with cls._async_session_context() as ctx:
            data = ModelCacheData(cache_id=cls.get_cache_id(id_value), data=update_data)
            return await cls.Meta.cache_protocol.cache_one(ctx, data)

    async def save(self, ttl: Optional[timedelta] = None, **kwargs) -> bool:
        async with self.__class__._async_session_context() as ctx:
            if not ttl:
                await self._register_instances(ctx, [self])
            data = ModelCacheData(cache_id=self.get_cache_id(self.get_id_value()), data=self.get_serializable_values())
            return await self.Meta.cache_protocol.cache_one(ctx, data, ttl=ttl, **kwargs)

    async def load(self, field_names: List[str] = None) -> bool:
        field_names = set(field_names) if field_names else set()

        async with self.__class__._async_session_context() as ctx:
            exists, data = await self.Meta.cache_protocol.load_one(ctx, self.get_cache_key(self.get_id_value()))
            if exists:
                self._construct(data, self, field_names)
                await self._build_rela(ctx, self, load_fields=field_names)
                return True
        return False

    @classmethod
    async def save_all(cls, objects: List['NoSqlModel'],
                       ttl: Optional[timedelta] = None, **kwargs) -> bool:
        async with cls._async_session_context() as ctx:
            objects_data = [
                ModelCacheData(cache_id=cls.get_cache_id(o.get_id_value()), data=o.get_serializable_values())
                for o in objects
            ]
            if res := await cls.Meta.cache_protocol.cache_many(ctx, objects_data, ttl, **kwargs):
                if not ttl:
                    await cls._register_instances(ctx, objects)
            return res

    async def delete(self) -> bool:
        async with self.__class__._async_session_context() as ctx:
            id_value = self.get_id_value()
            await self._unregister_indices(ctx, [id_value])
            return await self.Meta.cache_protocol.drop_one(ctx, self.get_cache_key(id_value))

    @classmethod
    async def delete_all(cls, indices: List[Any] = None) -> bool:
        async with cls._async_session_context() as ctx:
            await cls._unregister_indices(ctx, indices)
            res = await cls.Meta.cache_protocol.drop_many(ctx, indices)
            return res > 0

    @classmethod
    async def delete_by_id(cls, id_value: Any) -> bool:
        async with cls._async_session_context() as ctx:
            await cls._unregister_indices(ctx, [id_value])
            return await cls.Meta.cache_protocol.drop_one(ctx, cls.get_cache_key(id_value))


class NoSqlProxyModel(NoSqlModel):
    class Meta(NoSqlModel.Meta):
        sql_model: Model
        ttl_enabled: bool = False

    @classmethod
    def _fields_cache_key(cls):
        return cls.Meta.sql_model._fields_cache_key()

    @classmethod
    def get_cache_key(cls, id_value: Any) -> str:
        return f'{cls.get_model_cache_key()}:{cls.get_cache_id(id_value)}'

    #
    # Retrieve operations
    #

    @classmethod
    async def get(cls, id_value: Any, **kwargs) -> Optional['NoSqlProxyModel']:
        if not kwargs.pop('no_cache', False):
            if instance := await super().get(id_value, **kwargs):
                return instance
        instance = await cls.Meta.sql_model.get(id_value, **kwargs)
        if instance:
            instance = cls._construct(instance.get_values())
            await super(NoSqlProxyModel, instance).save(**kwargs)
        return instance

    @classmethod
    async def get_any(cls, indices: List[Any], **kwargs) -> List[Optional['NoSqlProxyModel']]:
        if not kwargs.pop('no_cache', False):
            if objects := await super().get_any(indices, **kwargs):
                return objects
        objects = await cls.Meta.sql_model.get_any(indices, **kwargs)
        if objects:
            objects = [cls._construct(o.get_values()) for o in objects]
            await super().save_all(objects, **kwargs)
        return objects

    @classmethod
    async def exists(cls, id_value: Any, **kwargs) -> bool:
        if not kwargs.pop('no_cache', False):
            return await super().exists(id_value, **kwargs)
        return await cls.Meta.sql_model.exists(id_value, **kwargs)

    #
    # Modification operations
    #

    async def save(self, ttl: Optional[timedelta] = None, **kwargs) -> bool:
        if res := await super().save(ttl, **kwargs):
            sql_instance = self.Meta.sql_model._construct(self.get_values())
            await sql_instance.save(**kwargs)
        return res

    async def load(self, field_names: List[str] = None) -> bool:
        if res := await super().load(field_names):
            return res
        sql_instance = self.Meta.sql_model._construct(self.get_values())
        if res := await sql_instance.load(field_names):
            self._construct(sql_instance.get_values(), instance=self)
        return res

    @classmethod
    async def save_all(cls, objects: List['NoSqlProxyModel'],
                       ttl: Optional[timedelta] = None, **kwargs) -> bool:
        if res := await super().save_all(objects):
            sql_objects = [cls.Meta.sql_model._construct(o.get_values()) for o in objects]
            return await cls.Meta.sql_model.save_all(sql_objects, ttl, **kwargs)
        return res

    async def delete(self) -> bool:
        if res := await super().delete():
            return await self.Meta.sql_model.delete_by_id(self.get_id_value())
        return res

    @classmethod
    async def delete_all(cls, indices: List[Any] = None) -> bool:
        if res := await super().delete_all(indices):
            return await cls.Meta.sql_model.delete_all(indices)
        return res



from collections import defaultdict
from typing import Protocol, List, Any, Optional

from .base import ModelProtocol
from ..context import ContextData


class ModelLocalCacheMixin(Protocol):
    @classmethod
    def _add_to_local_cache(cls, ctx: ContextData, instances: List[ModelProtocol]):
        if not hasattr(ctx, 'cache'):
            ctx.cache = defaultdict(lambda: defaultdict(dict))
        for instance in instances:
            if instance is not None:
                ctx.cache[cls][instance.get_id_value()] = instance

    @classmethod
    def _get_from_local_cache(cls, ctx: ContextData, indices: List[Any]) -> List[Optional[ModelProtocol]]:
        if not hasattr(ctx, 'cache'):
            ctx.cache = defaultdict(lambda: defaultdict(dict))

        return [ctx.cache[cls].get(id_value) for id_value in indices]

    @classmethod
    def _drop_from_local_cache(cls, ctx: ContextData, indices: List[Any] = None) -> List[bool]:
        if not hasattr(ctx, 'cache'):
            ctx.cache = defaultdict(lambda: defaultdict(dict))

        if indices is None:
            ctx.cache[cls] = defaultdict(dict)
            return [True]

        return [ctx.cache[cls].pop(id_value, None) is None for id_value in indices]

    @classmethod
    def _has_in_local_cache(cls, ctx: ContextData, indices: List[Any]) -> List[bool]:
        if not hasattr(ctx, 'cache'):
            ctx.cache = defaultdict(lambda: defaultdict(dict))

        return [id_value in ctx.cache[cls] for id_value in indices]

    # noinspection PyProtectedMember
    @classmethod
    async def _register_instances(cls, ctx: ContextData, objects: List[ModelProtocol]):
        await super()._register_instances(ctx, objects)
        cls._add_to_local_cache(ctx, objects)

    # noinspection PyProtectedMember
    @classmethod
    async def _unregister_indices(cls, ctx: ContextData, indices: List[Any] = None):
        cls._drop_from_local_cache(ctx, indices)
        await super()._unregister_indices(ctx, indices)



import asyncio
import random
import time
from multiprocessing import Array, Process, Value

import pytest
from dependency_injector import containers, providers

from dropland.blocks.redis import USE_REDIS
from dropland.blocks.scheduler import USE_SCHEDULER
from tests import REDIS_URI, SQLITE_URI

pytestmark = pytest.mark.skipif(not USE_SCHEDULER, reason='For scheduler only')

if USE_SCHEDULER:
    from dropland.blocks.scheduler.application import SchedulerApplication
    from dropland.blocks.scheduler.containers import SimpleSchedulerContainer, SchedulerContainer
    from dropland.blocks.scheduler.engine import EngineConfig
    default_scheduler_module = SimpleSchedulerContainer()


def test_create_engine():
    engine_factory = default_scheduler_module.engine_factory()

    assert not engine_factory.get_engine('')

    if USE_REDIS:
        config = EngineConfig(sql_url=SQLITE_URI, redis_url=REDIS_URI)
    else:
        config = EngineConfig(sql_url=SQLITE_URI)

    assert not engine_factory.get_engine('dropland')
    assert engine_factory.create_engine('dropland', config)
    assert engine_factory.get_engine('dropland')

    assert engine_factory.get_engine_names() == ['dropland']

    assert engine_factory.create_engine('dropland-remote', config)
    assert engine_factory.get_engine('dropland')
    assert engine_factory.get_engine('dropland-remote')

    assert engine_factory.get_engine_names() == ['dropland', 'dropland-remote']


job_called = 0
job_call_args = []
process_job_called = Value('i', 0)
process_job_call_args = Array('i', 2)
SLEEP_DURATION = 0.01


def task_fn(a):
    global job_called, job_call_args
    job_called += 1
    job_call_args.append(a)


async def async_task_fn(a):
    global job_called, job_call_args
    job_called += 1
    job_call_args.append(a)


def process_task_fn(a):
    global process_job_called, process_job_call_args
    process_job_call_args[process_job_called.value] = a
    process_job_called.value += 1
    time.sleep(SLEEP_DURATION)


async def async_process_task_fn(a):
    global process_job_called, process_job_call_args
    process_job_call_args[process_job_called.value] = a
    process_job_called.value += 1
    await asyncio.sleep(SLEEP_DURATION)


_current_app = None


def create_scheduler_app(with_remote: bool = False):
    settings = {
        'name': str(random.randint(0, 1000)),
        'engine_config': {
            'sql_url': SQLITE_URI,
            'create_remote_engine': with_remote,
            'task_host': 'localhost',
            'task_port': 3000,
        }
    }

    if USE_REDIS:
        settings['engine_config']['redis_url'] = REDIS_URI

    app_container = containers.DynamicContainer()
    app_container.blocks = providers.Aggregate(
        scheduler=providers.Factory(SchedulerContainer, config=settings)
    )

    app = SchedulerApplication(app_container, settings['name'], settings['engine_config'])
    app.path = 'tests.test_scheduler:_current_app'
    app_container.instance = providers.Object(app)

    global _current_app
    _current_app = app
    return app


@pytest.fixture(scope='function')
def scheduler_app():
    return create_scheduler_app()


@pytest.fixture(scope='function')
def remote_scheduler_app():
    return create_scheduler_app(True)


@pytest.mark.parametrize('target', (task_fn, async_task_fn), ids=['SyncTarget', 'AsyncTarget'])
def test_sync_run_tasks_as_local(target, scheduler_app):
    global job_called, job_call_args
    job_called, job_call_args = 0, []

    scheduler = scheduler_app.scheduler
    scheduler.start()
    scheduler.add_job(target, args=['123'], trigger='date')

    while job_called == 0:
        time.sleep(SLEEP_DURATION)

    scheduler.add_job(target, args=['456'], trigger='date', jobstore='sql')

    while job_called == 1:
        time.sleep(SLEEP_DURATION)

    if USE_REDIS:
        scheduler.add_job(target, args=['789'], trigger='date', jobstore='redis')

        while job_called == 2:
            time.sleep(SLEEP_DURATION)

    scheduler.shutdown()

    if USE_REDIS:
        assert job_called == 3
        assert job_call_args == ['123', '456', '789']
    else:
        assert job_called == 2
        assert job_call_args == ['123', '456']


@pytest.mark.asyncio
@pytest.mark.parametrize('target', (task_fn, async_task_fn), ids=['SyncTarget', 'AsyncTarget'])
async def test_async_run_tasks_as_local(target, scheduler_app):
    global job_called, job_call_args
    job_called, job_call_args = 0, []

    scheduler = scheduler_app.scheduler
    scheduler.start()
    scheduler.add_job(target, args=['123'], trigger='date')

    while job_called == 0:
        await asyncio.sleep(SLEEP_DURATION)

    scheduler.add_job(target, args=['456'], trigger='date', jobstore='sql')

    while job_called == 1:
        await asyncio.sleep(SLEEP_DURATION)

    if USE_REDIS:
        scheduler.add_job(target, args=['789'], trigger='date', jobstore='redis')

        while job_called == 2:
            await asyncio.sleep(SLEEP_DURATION)

    scheduler.shutdown()

    if USE_REDIS:
        assert job_called == 3
        assert job_call_args == ['123', '456', '789']
    else:
        assert job_called == 2
        assert job_call_args == ['123', '456']


@pytest.mark.parametrize('target', (process_task_fn, async_process_task_fn), ids=['SyncTarget', 'AsyncTarget'])
def test_sync_run_tasks_as_interprocess(target, scheduler_app):
    global process_job_called, process_job_call_args
    process_job_called, process_job_call_args = Value('i', 0), Array('i', 2)

    scheduler = scheduler_app.scheduler
    scheduler.start()
    scheduler.add_job(target, args=[1], trigger='date', jobstore='sql', executor='process')

    while process_job_called.value == 0:
        time.sleep(SLEEP_DURATION)

    if USE_REDIS:
        scheduler.add_job(target, args=[2], trigger='date', jobstore='redis', executor='process')

        while process_job_called.value == 1:
            time.sleep(SLEEP_DURATION)

    scheduler.shutdown()

    if USE_REDIS:
        assert process_job_called.value == 2
        assert process_job_call_args[:] == [1, 2]
    else:
        assert process_job_called.value == 1
        assert process_job_call_args[:] == [1, 0]


@pytest.mark.asyncio
@pytest.mark.parametrize('target', (process_task_fn, async_process_task_fn), ids=['SyncTarget', 'AsyncTarget'])
async def test_async_run_tasks_as_interprocess(target, scheduler_app):
    global process_job_called, process_job_call_args
    process_job_called, process_job_call_args = Value('i', 0), Array('i', 2)

    scheduler = scheduler_app.scheduler
    scheduler.start()
    scheduler.add_job(target, args=[3], trigger='date', jobstore='sql', executor='process')

    while process_job_called.value == 0:
        await asyncio.sleep(SLEEP_DURATION)

    if USE_REDIS:
        scheduler.add_job(target, args=[4], trigger='date', jobstore='redis', executor='process')

        while process_job_called.value == 1:
            await asyncio.sleep(SLEEP_DURATION)

    scheduler.shutdown()

    if USE_REDIS:
        assert process_job_called.value == 2
        assert process_job_call_args[:] == [3, 4]
    else:
        assert process_job_called.value == 1
        assert process_job_call_args[:] == [3, 0]


@pytest.mark.parametrize('target', (process_task_fn, async_process_task_fn), ids=['SyncTarget', 'AsyncTarget'])
@pytest.mark.parametrize('executor', ('default', 'process'), ids=['Default', 'Process'])
def test_run_tasks_as_remote(target, executor: str, remote_scheduler_app):
    global process_job_called, process_job_call_args
    process_job_called, process_job_call_args = Value('i', 0), Array('i', 2)

    from dropland.blocks.scheduler.application import run_service
    p = Process(target=run_service)
    p.start()
    time.sleep(SLEEP_DURATION * 10)

    remote_scheduler = remote_scheduler_app.scheduler
    remote_scheduler.start()
    remote_scheduler.add_job(target, args=[5], trigger='date', jobstore='sql', executor=executor)

    while process_job_called.value == 0:
        time.sleep(SLEEP_DURATION)

    if USE_REDIS:
        remote_scheduler.add_job(target, args=[6], trigger='date', jobstore='redis', executor=executor)

        while process_job_called.value == 1:
            time.sleep(SLEEP_DURATION)

    remote_scheduler.shutdown()
    time.sleep(SLEEP_DURATION * 10)
    p.terminate()
    p.join()

    if USE_REDIS:
        assert process_job_called.value == 2
        assert process_job_call_args[:] == [5, 6]
    else:
        assert process_job_called.value == 1
        assert process_job_call_args[:] == [5, 0]


@pytest.mark.parametrize('target', (process_task_fn, async_process_task_fn), ids=['SyncTarget', 'AsyncTarget'])
@pytest.mark.parametrize('executor', ('default', 'process'), ids=['Default', 'Process'])
@pytest.mark.parametrize('remote', (False, True), ids=['Local', 'Remote'])
def test_scheduled_job(target, executor: str, remote: bool):
    global process_job_called, process_job_call_args
    process_job_called, process_job_call_args = Value('i', 0), Array('i', 2)

    if remote:
        from dropland.blocks.scheduler.application import run_service
        p = Process(target=run_service)
        p.start()

    scheduler_app = create_scheduler_app(remote)
    scheduler = scheduler_app.scheduler
    assert scheduler
    scheduler.start()

    scheduler.scheduled_job(args=[7], trigger='date', jobstore='sql', executor=executor)(target)

    while process_job_called.value == 0:
        time.sleep(SLEEP_DURATION)

    if USE_REDIS:
        scheduler.scheduled_job(args=[8], trigger='date', jobstore='redis', executor=executor)(target)

        while process_job_called.value == 1:
            time.sleep(SLEEP_DURATION)

    scheduler.shutdown()
    time.sleep(SLEEP_DURATION * 10)

    if remote:
        # noinspection PyUnboundLocalVariable
        p.terminate()
        p.join()

    if USE_REDIS:
        assert process_job_called.value == 2
        assert process_job_call_args[:] == [7, 8]
    else:
        assert process_job_called.value == 1
        assert process_job_call_args[:] == [7, 0]



from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional, Protocol, runtime_checkable

from dependency_injector.containers import Container


@runtime_checkable
class Block(Protocol):
    pass


class ResourceBlock(Block):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._initialized: bool = False

    @property
    def initialized(self) -> bool:
        return self._initialized

    def sync_startup(self, application=None, service=None, module=None, *args, **kwargs):
        pass

    def sync_shutdown(self, application=None, service=None, module=None, *args, **kwargs):
        pass

    async def startup(self, application=None, service=None, module=None, *args, **kwargs):
        pass

    async def shutdown(self, application=None, service=None, module=None, *args, **kwargs):
        pass


class SessionResourceBlock(ResourceBlock):
    def sync_session_begin(self, application=None, service=None, module=None, *args, **kwargs):
        pass

    def sync_session_finish(self, application=None, service=None, module=None, *args, **kwargs):
        pass

    async def session_begin(self, application=None, service=None, module=None, *args, **kwargs):
        pass

    async def session_finish(self, application=None, service=None, module=None, *args, **kwargs):
        pass


class ContainerBlock(Block):
    def __init__(self, container: Container, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._container = container

    @property
    def container(self) -> Container:
        return self._container


class Module(ResourceBlock, ABC):
    name: str
    resources: List[ResourceBlock]

    @abstractmethod
    def get_resource(self, name) -> Optional[ResourceBlock]:
        ...


class Service(Module, ABC):
    modules: List[Module]

    @abstractmethod
    def get_module(self, name) -> Optional[Module]:
        ...


class Application(Service, ABC):
    services: List[Service]

    @abstractmethod
    def get_service(self, name) -> Optional[Service]:
        ...

    @property
    @abstractmethod
    def version(self) -> str:
        ...

    @property
    @abstractmethod
    def debug(self) -> bool:
        ...

    @abstractmethod
    def get_cwd(self) -> Path:
        ...

    @abstractmethod
    def sync_load_modules(self, *args, **kwargs):
        ...

    @abstractmethod
    async def load_modules(self, *args, **kwargs):
        ...



import contextlib
from datetime import datetime, timedelta
from typing import Any, Callable, Optional, Type

import rpyc
from apscheduler.util import obj_to_ref, undefined
from contextvars import ContextVar
from dropland.log import tr, worker_logger
from rpyc.core.protocol import Connection as RemoteConnection


# noinspection PyProtectedMember
def register_exception(exc: Type[Exception]):
    class_key = f'{exc.__module__}.{exc.__name__}'
    rpyc.core.vinegar._generic_exceptions_cache[class_key] = exc


class RemoteScheduler(object):
    def __init__(self, host: str, port: int, num_connect_attempts: int = 10, rpc_timeout: Optional[timedelta] = None):
        self._host = host
        self._port = port
        self._conn: ContextVar[RemoteConnection] = ContextVar('_conn')
        self._config = rpyc.core.protocol.DEFAULT_CONFIG.copy()
        self._config.update({'allow_public_attrs': True, 'allow_pickle': True})
        self._num_connect_attempts = max(1, num_connect_attempts)
        if rpc_timeout:
            self._config['sync_request_timeout'] = rpc_timeout.total_seconds()

    @contextlib.contextmanager
    def _ensure_connection(self):
        conn = self._conn.get(None)
        if conn:
            yield conn
        else:
            with rpyc.connect(self._host, self._port, config=self._config, keepalive=True) as conn:
                worker_logger.debug(tr('dropland.blocks.scheduler.service.rpc-connected')
                                    .format(host=self._host, port=self._port))

                token = self._conn.set(conn)
                try:
                    yield conn
                finally:
                    if self._conn.get(None) is conn:
                        self._conn.reset(token)
                        worker_logger.debug(tr('dropland.blocks.scheduler.service.rpc-disconnected')
                                            .format(host=self._host, port=self._port))

    def _execute(self, func: Callable[[RemoteConnection], Any]):
        attempt_num = self._num_connect_attempts

        while attempt_num > 0:
            try:
                with self._ensure_connection() as conn:
                    return func(conn)
            except (EOFError, OSError) as e:
                worker_logger.error(tr('dropland.blocks.scheduler.service.rpc-exception')
                                    .format(attempt_num=attempt_num, e=e))
                attempt_num -= 1

        return None

    def _prepare_args(self, args: tuple) -> tuple:
        result = []
        for value in args:
            if isinstance(value, datetime):
                value = value.strftime('%Y-%m-%dT%H:%M:%S.%f')
            elif isinstance(value, type(undefined)):
                value = undefined
            elif isinstance(value, Callable):
                value = obj_to_ref(value)
            result.append(value)
        return tuple(result)

    def _prepare_kwargs(self, args: dict) -> dict:
        for k, value in args.items():
            if isinstance(value, datetime):
                args[k] = value.strftime('%Y-%m-%dT%H:%M:%S.%f')
            elif isinstance(value, type(undefined)):
                args[k] = undefined
            elif isinstance(value, Callable):
                args[k] = obj_to_ref(value)
        return args

    def __getattr__(self, item):
        res = self._execute(lambda conn: getattr(conn.root, item))
        if not isinstance(res, Callable):
            return res

        def wrapped(*args, **kwargs):
            def _call(conn: RemoteConnection):
                func = getattr(conn.root, item)
                return func(*self._prepare_args(args), **self._prepare_kwargs(kwargs)) if func else None

            return self._execute(_call)

        return wrapped



import asyncio
import contextlib
from dataclasses import dataclass
from typing import Any, Optional, List, Dict, Union, Set

from sqlalchemy.sql import ClauseElement, ColumnCollection

from dropland.blocks.sql.base import SqlStorageType
from .cache import MethodCache
from .model import Model, CreateSchemaType, UpdateSchemaType
from .serializable import SerializableModel


@dataclass
class SqlTableInfo:
    db_type: SqlStorageType
    is_async: bool


class SqlModel(Model):
    @classmethod
    def get_table_info(cls) -> SqlTableInfo:
        ...

    @classmethod
    @contextlib.contextmanager
    def _session_context(cls, begin_tx: bool = False, autocommit: bool = False, session=None):
        ...

    @classmethod
    @contextlib.asynccontextmanager
    async def _async_session_context(
            cls, begin_tx: bool = False, autocommit: bool = False, session=None):
        ...

    @classmethod
    def query_for_select(cls, include_rela: Optional[Set[str]] = None,
                         exclude_rela: Optional[Set[str]] = None, **kwargs):
        ...

    @classmethod
    def query_for_update(cls, **kwargs):
        ...

    @classmethod
    def query_for_delete(cls, **kwargs):
        ...

    @classmethod
    def get_columns(cls) -> ColumnCollection:
        ...

    @classmethod
    def get_relationships(cls) -> Dict[str, Any]:
        ...

    @classmethod
    def query_options(cls, include: Optional[Set[str]] = None, exclude: Optional[Set[str]] = None):
        ...

    @classmethod
    def prepare_for_create(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        return data

    @classmethod
    def prepare_for_update(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        return data

    @classmethod
    def prepare_for_delete(cls, obj: 'SqlModel') -> 'SqlModel':
        return obj

    #
    # Retrieve operations
    #

    @classmethod
    async def get(cls, id_value: Any, **kwargs) -> Optional['SqlModel']:
        ...

    @classmethod
    async def get_any(cls, indices: List[Any], **kwargs) -> List[Optional['SqlModel']]:
        ...

    @classmethod
    async def list(
        cls, filters: Optional[List[Any]] = None, sorting: Optional[List[Any]] = None,
            skip: int = 0, limit: int = 0, params: Dict[str, Any] = None, **kwargs) -> List['SqlModel']:
        ...

    @classmethod
    async def count(cls, filters: Optional[List[Any]] = None, params: Dict[str, Any] = None, **kwargs) -> int:
        ...

    @classmethod
    async def exists(cls, id_value: Any, **kwargs) -> bool:
        ...

    @classmethod
    async def exists_by(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs) -> bool:
        ...

    #
    # Modification operations
    #

    @classmethod
    async def create(cls, data: Union[CreateSchemaType, Dict[str, Any]], **kwargs) -> Optional['SqlModel']:
        ...

    async def update(self, data: Union[UpdateSchemaType, Dict[str, Any]]) -> bool:
        ...

    @classmethod
    async def update_by_id(cls, id_value: Any, data: Union[UpdateSchemaType, Dict[str, Any]]) -> bool:
        ...

    @classmethod
    async def update_by(cls, filters: List[Any],
                        data: Union[UpdateSchemaType, Dict[str, Any]],
                        /, params: Dict[str, Any] = None, **kwargs) -> int:
        ...

    async def save(self, updated_fields: List[str] = None, **kwargs) -> bool:
        ...

    async def load(self, field_names: List[str] = None) -> bool:
        ...

    @classmethod
    async def save_all(cls, objects: List['SqlModel'], *args, **kwargs) -> bool:
        ...

    async def delete(self) -> bool:
        ...

    @classmethod
    async def delete_all(cls, indices: List[Any] = None) -> bool:
        ...

    @classmethod
    async def delete_by_id(cls, id_value: Any) -> bool:
        ...

    @classmethod
    async def delete_by(cls, filters: List[Any], /, params: Dict[str, Any] = None, **kwargs) -> int:
        ...

    #
    # Query operations
    #

    @classmethod
    def query_get(cls, id_value: Any, query=None, **kwargs):
        ...

    @classmethod
    def query_any(cls, indices: List[Any], query=None, **kwargs):
        ...

    @classmethod
    def query_list(cls, filters: List[Any], sorting: List[Any],
                   skip: int = 0, limit: int = 0, params: Dict[str, Any] = None, **kwargs):
        ...

    @classmethod
    def query_count(cls, filters: Optional[List[Any]] = None, params: Dict[str, Any] = None, **kwargs):
        ...

    @classmethod
    def query_exists(cls, id_value: Any, **kwargs):
        ...

    @classmethod
    def query_exists_by(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs):
        ...

    @classmethod
    def query_update(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs):
        ...

    @classmethod
    def query_delete(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs):
        ...


# noinspection PyProtectedMember
class CachedSqlModel(SqlModel, SerializableModel):
    class Meta(SqlModel.Meta):
        caches: List[MethodCache]

    @classmethod
    def has_cache(cls):
        return len(cls.Meta.caches) > 0

    #
    # Retrieve operations
    #

    @classmethod
    async def get(cls, id_value: Any, **kwargs) -> Optional['CachedSqlModel']:
        if kwargs.get('no_cache', False):
            return await super().get(id_value, **kwargs)

        exists: bool = False
        instance: Optional[CachedSqlModel] = None
        method_cache_key = MethodCache.cache_key_for('get', {'id': id_value, 'kw': kwargs})

        for cache in cls.Meta.caches:
            exists, data = await cache.get('get', cache_key=method_cache_key)
            if exists:
                if instance := cls._construct(data, **kwargs):
                    break

        if exists:
            async with cls._async_session_context() as ctx:
                await cls._register_instances(ctx, [instance])
                return await cls._build_rela(ctx, instance, **kwargs)

        # noinspection PyTypeChecker
        result = await super().get(id_value, no_cache=True, **kwargs)  # type: Optional[CachedSqlModel]
        serializable = result.get_serializable_values() if result is not None else None

        for cache in cls.Meta.caches:
            await cache.put('get', cache_key=method_cache_key, data=serializable)
        return result

    @classmethod
    async def get_any(cls, indices: List[Any], **kwargs) -> List[Optional['CachedSqlModel']]:
        if kwargs.get('no_cache', False):
            return await super().get_any(indices, **kwargs)

        exists: bool = False
        objects: Optional[List[CachedSqlModel]] = None
        method_cache_key = MethodCache.cache_key_for('get_any', {'indices': indices, 'kw': kwargs})

        for cache in cls.Meta.caches:
            exists, data = await cache.get('get_any', cache_key=method_cache_key)
            if exists:
                # noinspection PyTypeChecker
                if objects := cls._construct_list(data, **kwargs):
                    break

        if exists:
            async with cls._async_session_context() as ctx:
                await cls._register_instances(ctx, objects)
                # noinspection PyTypeChecker
                return await cls._build_rela_list(ctx, objects, **kwargs)

        # noinspection PyTypeChecker
        result = await super().get_any(indices, no_cache=True, **kwargs)  # type: List[Optional[CachedSqlModel]]
        serializable = [i.get_serializable_values() if i is not None else None for i in result]

        for cache in cls.Meta.caches:
            await cache.put('get_any', cache_key=method_cache_key, data=serializable)
        return result

    @classmethod
    async def list(
        cls, filters: Optional[List[Any]] = None, sorting: Optional[List[Any]] = None,
            skip: int = 0, limit: int = 0, params: Dict[str, Any] = None, **kwargs) -> List['CachedSqlModel']:
        if kwargs.get('no_cache', False):
            # noinspection PyTypeChecker
            return await super().list(filters, sorting, skip, limit, params, **kwargs)

        exists: bool = False
        objects: Optional[List[CachedSqlModel]] = None
        method_cache_key = MethodCache.cache_key_for('list', {
            'filters': cls._expr2string(filters, params),
            'sorting': cls._expr2string(sorting),
            'skip': skip, 'limit': limit, 'kw': kwargs
        })

        for cache in cls.Meta.caches:
            exists, data = await cache.get('list', cache_key=method_cache_key)
            if exists:
                # noinspection PyTypeChecker
                if objects := cls._construct_list(data, **kwargs):
                    break

        if exists:
            async with cls._async_session_context() as ctx:
                await cls._register_instances(ctx, objects)
                # noinspection PyTypeChecker
                return await cls._build_rela_list(ctx, objects, **kwargs)

        # noinspection PyTypeChecker
        result = await super().list(filters, sorting, skip, limit, params, no_cache=True, **kwargs)  # type: List['CachedSqlModel']
        serializable = [i.get_serializable_values() for i in result]

        for cache in cls.Meta.caches:
            await cache.put('list', cache_key=method_cache_key, data=serializable)
        return result

    @classmethod
    async def count(cls, filters: Optional[List[Any]] = None, params: Dict[str, Any] = None, **kwargs) -> int:
        if kwargs.get('no_cache', False):
            return await super().count(filters, params, **kwargs)

        method_cache_key = MethodCache.cache_key_for('count', {
            'filters': cls._expr2string(filters, params), 'kw': kwargs
        })

        for cache in cls.Meta.caches:
            exists, data = await cache.get('count', cache_key=method_cache_key)
            if exists:
                return data

        result = await super().count(filters, params, no_cache=True, **kwargs)

        for cache in cls.Meta.caches:
            await cache.put('count', cache_key=method_cache_key, data=result)
        return result

    @classmethod
    async def exists(cls, id_value: Any, **kwargs) -> bool:
        if kwargs.get('no_cache', False):
            return await super().exists(id_value, **kwargs)

        method_cache_key = MethodCache.cache_key_for('exists', {'id': id_value, 'kw': kwargs})

        for cache in cls.Meta.caches:
            exists, data = await cache.get('exists', cache_key=method_cache_key)
            if exists:
                return data

        result = await super().exists(id_value, no_cache=True, **kwargs)

        for cache in cls.Meta.caches:
            await cache.put('exists', cache_key=method_cache_key, data=result)
        return result

    @classmethod
    async def exists_by(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs) -> bool:
        if kwargs.get('no_cache', False):
            return await super().exists_by(filters, params, **kwargs)

        method_cache_key = MethodCache.cache_key_for('exists_by', {
            'filters': cls._expr2string(filters, params), 'kw': kwargs
        })

        for cache in cls.Meta.caches:
            exists, data = await cache.get('exists_by', cache_key=method_cache_key)
            if exists:
                return data

        result = await super().exists_by(filters, params, no_cache=True, **kwargs)

        for cache in cls.Meta.caches:
            await cache.put('exists_by', cache_key=method_cache_key, data=result)
        return result

    #
    # Modification operations
    #

    @classmethod
    async def create(cls, data: Union[CreateSchemaType, Dict[str, Any]], **kwargs) -> Optional['SqlModel']:
        if result := await super().create(data, **kwargs):
            await cls._drop_methods()
        return result

    async def update(self, data: Union[UpdateSchemaType, Dict[str, Any]]) -> bool:
        if result := await super().update(data):
            await self._drop_methods(['get', 'get_any', 'list'])
        return result

    @classmethod
    async def update_by_id(cls, id_value: Any, data: Union[UpdateSchemaType, Dict[str, Any]]) -> bool:
        if result := await super().update_by_id(id_value, data):
            await cls._drop_methods(['get', 'get_any', 'list'])
        return result

    @classmethod
    async def update_by(cls, filters: List[Any],
                        data: Union[UpdateSchemaType, Dict[str, Any]],
                        /, params: Dict[str, Any] = None, **kwargs) -> int:
        if result := await super().update_by(filters, data, params=params, **kwargs):
            await cls._drop_methods(['get', 'get_any', 'list'])
        return result

    async def delete(self) -> bool:
        if result := await super().delete():
            await self._drop_methods()
        return result

    @classmethod
    async def delete_all(cls, indices: List[Any] = None) -> bool:
        if result := await super().delete_all(indices):
            await cls._drop_methods()
        return result

    @classmethod
    async def delete_by_id(cls, id_value: Any) -> bool:
        if result := await super().delete_by_id(id_value):
            await cls._drop_methods()
        return result

    @classmethod
    async def delete_by(cls, filters: List[Any], /, params: Dict[str, Any] = None, **kwargs) -> int:
        if result := await super().delete_by(filters, params=params, **kwargs):
            await cls._drop_methods()
        return result

    async def save(self, updated_fields: List[str] = None, **kwargs) -> bool:
        if result := await super().save(updated_fields, **kwargs):
            await self._drop_methods(['get', 'get_any', 'list'])
        return result

    async def load(self, field_names: List[str] = None) -> bool:
        if result := await super().load(field_names):
            await self._drop_methods(['get', 'get_any', 'list'])
        return result

    @classmethod
    async def save_all(cls, objects: List['SqlModel'], *args, **kwargs) -> bool:
        if result := await super().save_all(objects, *args, **kwargs):
            await cls._drop_methods()
        return result

    #
    # Private
    #

    @classmethod
    async def _drop_methods(cls, methods: List[str] = None):
        method_names = ('get', 'get_any', 'list', 'count', 'exists', 'exists_by')
        methods = set(methods) if methods else set(method_names)

        await asyncio.gather(
            *(cache.drop_all([m for m in method_names if m in methods]) for cache in cls.Meta.caches)
        )

    @classmethod
    def _expr2string(cls, expr, params: dict = None):
        if expr is None:
            return None
        elif isinstance(expr, list):
            res = [cls._expr2string(i, params) for i in expr]
            return ','.join(res)
        elif isinstance(expr, ClauseElement):
            if params:
                expr = expr.params(params)
            compiled = expr.compile(compile_kwargs={'render_postcompile': True})
            res = [compiled.string]
            for k, v in compiled.params.items():
                if params and k in params:
                    res.append(str(params[k]))
                else:
                    res.append(str(v))
            return '$'.join(res)

        return str(expr)
